import numpy as np
import base64
import io
from PIL import Image

from . import BaseAnnotationDefinition

from .box import Box
from .polygon import Polygon


class Segmentation(BaseAnnotationDefinition):
    """
    Segmentation annotation object
    """
    type = "binary"

    def __init__(self, geo: np.ndarray, label: str, attributes=None, description=None, color=None):
        super().__init__(description=description, attributes=attributes)
        self._geo = geo
        self._coordinates = None
        self.label = label
        self._color = color

    @property
    def geo(self) -> np.ndarray:
        if self._geo is None:
            self._geo = self.from_coordinates(self._coordinates)
            if self._color is None:
                color = None
                fill_coordinates = self._geo.nonzero()
                if len(fill_coordinates) > 0 and len(fill_coordinates[0]) > 0 and len(fill_coordinates[1]) > 0:
                    color = self._geo[fill_coordinates[0][0]][fill_coordinates[1][0]]
                self._color = color
            self._geo = (self._geo[:, :, 3] > 127).astype(float)
        return self._geo

    @geo.setter
    def geo(self, geo: np.ndarray):
        self._geo = geo
        self._coordinates = None

    @property
    def x(self):
        return

    @property
    def y(self):
        return

    @property
    def pts(self):
        return np.where(self.geo > 0)

    @property
    def left(self):
        left = 0
        if len(self.pts[1]) > 0:
            left = np.min(self.pts[1])
        return left

    @property
    def top(self):
        top = 0
        if len(self.pts[0]) > 0:
            top = np.min(self.pts[0])
        return top

    @property
    def right(self):
        right = 0
        if len(self.pts[1]) > 0:
            right = np.max(self.pts[1])
        return right

    @property
    def bottom(self):
        bottom = 0
        if len(self.pts[0]) > 0:
            bottom = np.max(self.pts[0])
        return bottom

    def show(self, image, thickness, with_text, height, width, annotation_format, color, alpha=1):
        """
        Show annotation as ndarray
        :param image: empty or image to draw on
        :param thickness:
        :param with_text: not required
        :param height: item height
        :param width: item width
        :param annotation_format: options: list(dl.ViewAnnotationOptions)
        :param color: color
        :param alpha: opacity value [0 1], default 1
        :return: ndarray
        """
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            raise ImportError('opencv not found. Must install to perform this function')

        if alpha != 1:
            overlay = image.copy()
        else:
            overlay = image
        if color is None:
            if self._color:
                color = self._color
            else:
                color = (255, 255, 255)
        # draw annotation
        overlay[np.where(self.geo)] = color

        if not isinstance(color, int) and len(color) == 4 and color[3] != 255:
            # add with opacity
            image = cv2.addWeighted(src1=overlay,
                                    alpha=alpha,
                                    src2=image,
                                    beta=1 - alpha,
                                    gamma=0)
        else:
            image = overlay

        if with_text:
            image = self.add_text_to_image(image=image, annotation=self)
        return image

    def to_coordinates(self, color=None):
        """
        Convert segmentation to coordinates
        `self._color` is ignored if `color` is provided
        
        :param color: color
        :return: coordinates
        """
        need_encode = False
        if color is not None and self._color is not None:
            # if input color is not the same as the annotation's color - need to re-encode
            if self._color != color:
                need_encode = True

        if need_encode or self._coordinates is None:
            if color is None:
                if self._color is not None:
                    color = self._color
                else:
                    color = (255, 255, 255)
            max_val = np.max(self.geo)
            if max_val > 1:
                self.geo = self.geo / max_val
            png_ann = np.stack((color[0] * self.geo,
                                color[1] * self.geo,
                                color[2] * self.geo,
                                255 * self.geo),
                               axis=2).astype(np.uint8)
            pil_img = Image.fromarray(png_ann)
            buff = io.BytesIO()
            pil_img.save(buff, format="PNG")
            new_image_string = base64.b64encode(buff.getvalue()).decode("utf-8")
            self._coordinates = "data:image/png;base64,%s" % new_image_string
        return self._coordinates

    def to_box(self):
        """

        :return: Box annotations list  to each separated  segmentation
        """
        polygons = Polygon.from_segmentation(mask=self.geo, label=self.label,
                                             attributes=self.attributes, max_instances=None)

        if not isinstance(polygons, list):
            polygons = [polygons]

        boxes = [Box(left=polygon.left,
                     top=polygon.top,
                     right=polygon.right,
                     bottom=polygon.bottom,
                     label=polygon.label,
                     attributes=polygon.attributes) for polygon in polygons]

        return boxes

    @classmethod
    def from_polygon(cls, geo, label, shape, attributes=None):
        """

        :param geo: list of x,y coordinates of the polygon ([[x,y],[x,y]...]
        :param label: annotation's label
        :param shape: image shape (h,w)
        :param attributes:
        :return:
        """
        try:
            import cv2
        except (ImportError, ModuleNotFoundError):
            raise ImportError('opencv not found. Must install to perform this function')

        thickness = -1

        # plot polygon on a blank mask with thickness -1 to fill the polyline
        mask = np.zeros(shape=shape, dtype=np.uint8)
        mask = cv2.drawContours(image=mask,
                                contours=[np.asarray(geo).astype('int')],
                                contourIdx=-1,
                                color=1,
                                thickness=thickness)

        return cls(
            geo=mask,
            label=label,
            attributes=attributes,
        )

    @staticmethod
    def from_coordinates(coordinates):
        if isinstance(coordinates, dict):
            data = coordinates["data"][22:]
        elif isinstance(coordinates, str):
            data = coordinates[22:]
        else:
            raise TypeError('unknown binary data type')
        decode = base64.b64decode(data)
        mask = np.array(Image.open(io.BytesIO(decode)))
        return mask

    @classmethod
    def from_json(cls, _json):
        if "coordinates" in _json:
            coordinates = _json["coordinates"]
        elif "data" in _json:
            coordinates = _json["data"]
        else:
            raise ValueError('can not find "coordinates" or "data" in annotation. id: {}'.format(_json["id"]))
        inst = cls(
            geo=None,
            label=_json["label"],
            attributes=_json.get("attributes", None),
            color=None
        )
        inst._coordinates = coordinates
        return inst



from .cookie import CookieIO


class CallsCounter:
    def __init__(self, filepath):
        self.io = CookieIO(filepath)
        self.state = 'off'
        self.number = 0
        self.load()

    def add(self):
        if self.state == 'on':
            self.load()
            self.number += 1
            self.save()

    def reset(self):
        self.number = 0
        self.save()

    def save(self):
        self.io.put('calls_counter', {'state': self.state,
                                      'number': self.number})

    def on(self):
        self.state = 'on'
        self.save()

    def off(self):
        self.state = 'off'
        self.save()

    def on_exit(self):
        self.save()

    def load(self):
        calls = self.io.get('calls_counter')
        if calls is None:
            # not set in cookie - set default
            calls = {'state': 'off',
                     'number': 0}
            self.io.put('calls_counter', calls)
        self.state = calls['state']
        self.number = calls['number']



from .base_annotation_definition import BaseAnnotationDefinition
from .box import Box
from .cube import Cube
from .cube_3d import Cube3d
from .note import Note, Message
from .classification import Classification
from .comparison import Comparison
from .ellipse import Ellipse
from .point import Point
from .polygon import Polygon
from .polyline import Polyline
from .segmentation import Segmentation
from .subtitle import Subtitle
from .description import Description
from .undefined_annotation import UndefinedAnnotationType
from .pose import Pose
from .text import Text
from .free_text import FreeText
from .ref_image import RefImage
from .gis import Gis, GisType



import os
import tempfile
from pathlib import Path
from .. import exceptions
from ..services import service_defaults


class PathUtils:
    """
    Utility class for path validation and sanitization to prevent path traversal attacks.
    """
    allowed_roots = [tempfile.gettempdir(), service_defaults.DATALOOP_PATH]
    
    @staticmethod
    def _contains_traversal(path: str) -> bool:
        """
        Check if path contains path traversal sequences.
        
        :param str path: Path to check
        :return: True if path contains traversal sequences
        :rtype: bool
        """
        if not path:
            return False
        
        # Normalize the path to handle different separators
        normalized = os.path.normpath(path)
        
        # Check for parent directory references
        parts = Path(normalized).parts
        if '..' in parts:
            return True
        
        # Check for encoded traversal sequences (evasion attempts)
        if '%2e%2e' in path.lower() or '..%2f' in path.lower() or '..%5c' in path.lower():
            return True
        
        return False
    
    @staticmethod
    def _is_within_base(resolved_path: str, base_path: str) -> bool:
        """
        Check if resolved_path is within base_path.
        
        :param str resolved_path: Absolute resolved path
        :param str base_path: Base directory path
        :return: True if resolved_path is within base_path
        :rtype: bool
        """
        try:
            resolved = os.path.abspath(os.path.normpath(resolved_path))
            base = os.path.abspath(os.path.normpath(base_path))
            
            # Get common path
            common = os.path.commonpath([resolved, base])
            return common == base
        except (ValueError, OSError):
            # On Windows, if paths are on different drives, commonpath raises ValueError
            return False
    
    @staticmethod
    def _is_allowed_path(resolved_path: str, base_path: str) -> bool:
        """
        Check if resolved_path is within base_path or any allowed_root.
        
        :param str resolved_path: Absolute resolved path
        :param str base_path: Base directory path
        :return: True if resolved_path is within base_path or any allowed_root
        :rtype: bool
        """
        for allowed_root in [base_path] + PathUtils.allowed_roots:
            if PathUtils._is_within_base(resolved_path, allowed_root):
                return True
        return False
    
    @staticmethod
    def validate_directory_name(name: str) -> str:
        """
        Validate a directory name to ensure it doesn't contain path traversal sequences.
        
        :param str name: Directory name to validate
        :return: Validated directory name
        :rtype: str
        :raises PlatformException: If name contains invalid characters or traversal sequences
        """
        if not name:
            raise exceptions.PlatformException(
                error='400',
                message='Directory name cannot be empty'
            )
        
        # Check for path separators
        if os.sep in name or (os.altsep and os.altsep in name):
            raise exceptions.PlatformException(
                error='400',
                message='Directory name cannot contain path separators'
            )
        
        # Check for traversal sequences
        if PathUtils._contains_traversal(name):
            raise exceptions.PlatformException(
                error='400',
                message='Directory name cannot contain path traversal sequences'
            )
        
        return name
    
    @staticmethod
    def _validate_single_path(path, base_path: str, must_exist: bool):
        """
        Internal method to validate a single path string.
        
        :param path: Path to validate (str or Path object)
        :param str base_path: Base directory to restrict path to
        :param bool must_exist: If True, path must exist
        :raises PlatformException: If path is invalid or contains traversal sequences
        """
        # Convert Path objects to strings
        if isinstance(path, Path):
            path = str(path)
        if isinstance(base_path, Path):
            base_path = str(base_path)
        
        # Skip validation if not a string
        if not isinstance(path, str):
            return
        
        # Skip validation for URLs and external paths
        if path.startswith(('http://', 'https://', 'external://')):
            return
        
        # Empty string check
        if not path:
            raise exceptions.PlatformException(
                error='400',
                message='Path cannot be empty'
            )
        
        # Check for traversal sequences in the original path
        if PathUtils._contains_traversal(path):
            raise exceptions.PlatformException(
                error='400',
                message='Path contains invalid traversal sequences'
            )
        
        # Resolve path (absolute paths allowed if within base_path)
        if os.path.isabs(path):
            resolved = os.path.abspath(os.path.normpath(path))
        else:
            resolved = os.path.abspath(os.path.normpath(os.path.join(base_path, path)))
        
        # Reject if path is outside base_path or allowed_roots
        if not PathUtils._is_allowed_path(resolved, base_path):
            raise exceptions.PlatformException(
                error='400',
                message='Path resolves outside allowed directory'
            )
        
        # Check if path must exist
        if must_exist and not os.path.exists(resolved):
            raise exceptions.PlatformException(
                error='404',
                message='Path does not exist: {}'.format(path)
            )
    
    @staticmethod
    def validate_paths(paths, base_path = None, must_exist: bool = False):
        """
        Validate file or directory paths against path traversal attacks.
        Accepts a list of paths and validates each one.
        Skips validation if path is None or not a string.
        Skips validation for URLs (http://, https://) and external paths (external://).
        
        :param paths: Path(s) to validate - can be str, Path, list of str/Path, or None
        :param base_path: Optional base directory to restrict path to (str or Path). If None, uses current working directory
        :param bool must_exist: If True, path must exist
        :raises PlatformException: If any path is invalid or contains traversal sequences
        """
        # Handle None - skip validation
        if paths is None:
            return
        
        # Convert base_path Path object to string
        if isinstance(base_path, Path):
            base_path = str(base_path)
        
        # Resolve base_path
        if base_path is None:
            base_path = os.getcwd()
        
        # Handle list of paths
        if isinstance(paths, list):
            for path in paths:
                PathUtils._validate_single_path(path, base_path, must_exist)
        else:
            # Single path
            PathUtils._validate_single_path(paths, base_path, must_exist)
    
    @staticmethod
    def validate_file_path(file_path, base_path = None, must_exist: bool = True):
        """
        Validate a file path against path traversal attacks.
        
        :param file_path: File path to validate (str or Path object)
        :param base_path: Optional base directory to restrict path to (str or Path). If None, uses current working directory
        :param bool must_exist: If True, file must exist (default: True)
        :raises PlatformException: If path is invalid, contains traversal sequences, or is not a file
        """
        # Convert Path objects to strings
        if isinstance(file_path, Path):
            file_path = str(file_path)
        if isinstance(base_path, Path):
            base_path = str(base_path)
        
        PathUtils.validate_paths(file_path, base_path=base_path, must_exist=must_exist)
        
        if must_exist and isinstance(file_path, str) and not file_path.startswith(('http://', 'https://', 'external://')):
            # Resolve path to check if it's a file
            if base_path is None:
                base_path = os.getcwd()
            if os.path.isabs(file_path):
                resolved = os.path.abspath(os.path.normpath(file_path))
            else:
                resolved = os.path.abspath(os.path.normpath(os.path.join(base_path, file_path)))
            
            if not os.path.isfile(resolved):
                raise exceptions.PlatformException(
                    error='400',
                    message='Path is not a file: {}'.format(file_path)
                )
    
    @staticmethod
    def validate_directory_path(dir_path, base_path = None, must_exist: bool = True):
        """
        Validate a directory path against path traversal attacks.
        
        :param dir_path: Directory path to validate (str or Path object)
        :param base_path: Optional base directory to restrict path to (str or Path). If None, uses current working directory
        :param bool must_exist: If True, directory must exist (default: True)
        :raises PlatformException: If path is invalid, contains traversal sequences, or is not a directory
        """
        # Convert Path objects to strings
        if isinstance(dir_path, Path):
            dir_path = str(dir_path)
        if isinstance(base_path, Path):
            base_path = str(base_path)
        
        PathUtils.validate_paths(dir_path, base_path=base_path, must_exist=must_exist)
        
        if must_exist and isinstance(dir_path, str) and not dir_path.startswith(('http://', 'https://', 'external://')):
            # Resolve path to check if it's a directory
            if base_path is None:
                base_path = os.getcwd()
            if os.path.isabs(dir_path):
                resolved = os.path.abspath(os.path.normpath(dir_path))
            else:
                resolved = os.path.abspath(os.path.normpath(os.path.join(base_path, dir_path)))
            
            if not os.path.isdir(resolved):
                raise exceptions.PlatformException(
                    error='400',
                    message='Path is not a directory: {}'.format(dir_path)
                )




from collections import namedtuple
from typing import Union
from enum import Enum
import traceback
import logging
import inspect
import typing
import json
import os

from .package_module import PackageModule
from .package_slot import PackageSlot
from .. import repositories, entities, exceptions
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class RequirementOperator(str, Enum):
    EQUAL = '==',
    GREATER_THAN = '>',
    LESS_THAN = '<',
    EQUAL_OR_LESS_THAN = '<=',
    EQUAL_OR_GREATER_THAN = '>='

    @staticmethod
    def keys():
        return [key.value for key in list(RequirementOperator)]


class PackageRequirement:

    def __init__(self, name: str, version: str = None, operator: str = None):
        self.name = name
        self.version = version

        valid_operators = RequirementOperator.keys()
        if operator is not None and operator not in valid_operators:
            raise Exception('Illegal operator: {}. Please select from: {}'.format(operator, valid_operators))

        self.operator = operator

    def to_json(self):
        _json = {'name': self.name}
        if self.version is not None:
            _json['version'] = self.version
        if self.operator is not None:
            _json['operator'] = self.operator
        return _json

    @classmethod
    def from_json(cls, _json: dict):
        return cls(**_json)


class Package(entities.DlEntity):
    """
    Package object
    """
    # platform
    id: str = entities.DlProperty(location=['id'], _type=str)
    url: str = entities.DlProperty(location=['url'], _type=str)
    name: str = entities.DlProperty(location=['name'], _type=str)
    version: str = entities.DlProperty(location=['version'], _type=str)
    created_at: str = entities.DlProperty(location=['createdAt'], _type=str)
    updated_at: str = entities.DlProperty(location=['updatedAt'], _type=str)
    project_id: str = entities.DlProperty(location=['projectId'], _type=str)
    creator: str = entities.DlProperty(location=['creator'], _type=str)
    type: str = entities.DlProperty(location=['type'], _type=str)
    metadata: dict = entities.DlProperty(location=['metadata'], _type=dict)
    ui_hooks: list = entities.DlProperty(location=['uiHooks'], _type=str)
    service_config: dict = entities.DlProperty(location=['serviceConfig'], _type=str)
    is_global: bool = entities.DlProperty(location=['global'], _type=str)

    codebase: typing.Any = entities.DlProperty(location=['codebase'], _kls='Codebase')
    modules: typing.List[PackageModule] = entities.DlProperty(location=['modules'], _kls='PackageModule')
    slots: typing.Union[typing.List[PackageSlot], None] = entities.DlProperty(location=['slots'],
                                                                              _kls='PackageSlot')
    requirements: typing.Union[typing.List[PackageRequirement], None] = entities.DlProperty(location=['requirements'],
                                                                                            _kls='PackageRequirement')

    # sdk
    _client_api: ApiClient
    _revisions = None
    __repositories = None
    _project = None

    def __repr__(self):
        # TODO need to move to DlEntity
        return "Package(id={id}, name={name}, creator={creator}, project_id={project_id}, type={type}, version={version})".format(
            id=self.id,
            name=self.name,
            version=self.version,
            type=self.type,
            project_id=self.project_id,
            creator=self.creator)

    @property
    def createdAt(self):
        return self.created_at

    @property
    def updatedAt(self):
        return self.updated_at

    @property
    def revisions(self):
        if self._revisions is None:
            self._revisions = self.packages.revisions(package=self)
        return self._revisions

    @property
    def platform_url(self):
        return self._client_api._get_resource_url("projects/{}/packages/{}/main".format(self.project.id, self.id))

    @property
    def codebase_id(self):
        if self.codebase is not None and self.codebase.type == entities.PackageCodebaseType.ITEM:
            return self.codebase.item_id
        return None

    @codebase_id.setter
    def codebase_id(self, item_id: str):
        self.codebase = entities.ItemCodebase(item_id=item_id)

    @staticmethod
    def _protected_from_json(_json, client_api, project=None, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error

        :param _json:  platform json
        :param client_api: ApiClient entity
        :return:
        """
        try:
            package = Package.from_json(_json=_json,
                                        client_api=client_api,
                                        project=project,
                                        is_fetched=is_fetched)
            status = True
        except Exception:
            package = traceback.format_exc()
            status = False
        return status, package

    @classmethod
    def from_json(cls, _json, client_api, project=None, is_fetched=True):
        """
        Turn platform representation of package into a package entity

        :param dict _json: platform representation of package
        :param dl.ApiClient client_api: ApiClient entity
        :param dtlpy.entities.project.Project project: project entity
        :param is_fetched: is Entity fetched from Platform
        :return: Package entity
        :rtype: dtlpy.entities.package.Package
        """
        if project is not None:
            if project.id != _json.get('projectId', None):
                logger.warning('Package has been fetched from a project that is not belong to it')
                project = None
        # Entity
        inst = cls(_dict=_json)
        # Platform
        inst._project = project
        inst._client_api = client_api
        inst.is_fetched = is_fetched

        return inst

    def to_json(self):
        """
        Turn Package entity into a platform representation of Package

        :return: platform json of package
        :rtype: dict
        """
        _json = self._dict.copy()
        return _json

    ############
    # entities #
    ############
    @property
    def project(self):
        if self._project is None:
            self._project = self.projects.get(project_id=self.project_id, fetch=None)
        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project):
        assert isinstance(self._project, entities.Project), "Unknwon 'project' type: {}".format(type(project))
        self._project = project

    ################
    # repositories #
    ################
    @property
    def _repositories(self):
        if self.__repositories is None:
            reps = namedtuple('repositories',
                              field_names=['executions', 'services', 'projects', 'packages', 'artifacts', 'codebases',
                                           'models'])

            self.__repositories = reps(
                executions=repositories.Executions(client_api=self._client_api,
                                                   project=self._project),
                services=repositories.Services(client_api=self._client_api,
                                               package=self,
                                               project=self._project,
                                               project_id=self.project_id),
                projects=repositories.Projects(client_api=self._client_api),
                packages=repositories.Packages(client_api=self._client_api,
                                               project=self._project),
                artifacts=repositories.Artifacts(client_api=self._client_api,
                                                 project=self._project,
                                                 project_id=self.project_id,
                                                 package=self),
                codebases=repositories.Codebases(client_api=self._client_api, project=self._project,
                                                 project_id=self.project_id),
                models=repositories.Models(client_api=self._client_api,
                                           project=self._project,
                                           package=self,
                                           project_id=self.project_id)
            )
        return self.__repositories

    @property
    def executions(self):
        assert isinstance(self._repositories.executions, repositories.Executions)
        return self._repositories.executions

    @property
    def services(self):
        assert isinstance(self._repositories.services, repositories.Services)
        return self._repositories.services

    @property
    def projects(self):
        assert isinstance(self._repositories.projects, repositories.Projects)
        return self._repositories.projects

    @property
    def packages(self):
        assert isinstance(self._repositories.packages, repositories.Packages)
        return self._repositories.packages

    @property
    def codebases(self):
        assert isinstance(self._repositories.codebases, repositories.Codebases)
        return self._repositories.codebases

    @property
    def artifacts(self):
        assert isinstance(self._repositories.artifacts, repositories.Artifacts)
        return self._repositories.artifacts

    @property
    def models(self):
        assert isinstance(self._repositories.models, repositories.Models)
        return self._repositories.models

    ##############
    # properties #
    ##############
    @property
    def git_status(self):
        status = 'Git status unavailable'
        try:
            if self.codebase.type == entities.PackageCodebaseType.ITEM:
                if 'git' in self.codebase.item.metadata:
                    status = self.codebase.item.metadata['git'].get('status', status)
        except Exception:
            logging.debug('Error getting codebase')
        return status

    @property
    def git_log(self):
        log = 'Git log unavailable'
        try:
            if self.codebase.type == entities.PackageCodebaseType.ITEM:
                if 'git' in self.codebase.item.metadata:
                    log = self.codebase.item.metadata['git'].get('log', log)
        except Exception:
            logging.debug('Error getting codebase')
        return log

    ###########
    # methods #
    ###########
    def update(self):
        """
        Update Package changes to platform

        :return: Package entity
        """
        return self.packages.update(package=self)

    def deploy(self,
               service_name=None,
               revision=None,
               init_input=None,
               runtime=None,
               sdk_version=None,
               agent_versions=None,
               verify=True,
               bot=None,
               pod_type=None,
               module_name=None,
               run_execution_as_process=None,
               execution_timeout=None,
               drain_time=None,
               on_reset=None,
               max_attempts=None,
               force=False,
               secrets: list = None,
               **kwargs):
        """
        Deploy package

        :param str service_name: service name
        :param str revision: package revision - default=latest
        :param init_input: config to run at startup
        :param dict runtime: runtime resources
        :param str sdk_version:  - optional - string - sdk version
        :param dict agent_versions: - dictionary - - optional -versions of sdk, agent runner and agent proxy
        :param str bot: bot email
        :param str pod_type: pod type dl.InstanceCatalog
        :param bool verify: verify the inputs
        :param str module_name: module name
        :param bool run_execution_as_process: run execution as process
        :param int execution_timeout: execution timeout
        :param int drain_time: drain time
        :param str on_reset: on reset
        :param int max_attempts: Maximum execution retries in-case of a service reset
        :param bool force: optional - terminate old replicas immediately
        :param list secrets: list of the integrations ids
        :return: Service object
        :rtype: dtlpy.entities.service.Service

        **Example**:

        .. code-block:: python
        service: dl.Service = package.deploy(service_name=package_name,
                                             execution_timeout=3 * 60 * 60,
                                             module_name=module.name,
                                             runtime=dl.KubernetesRuntime(
                                                 concurrency=10,
                                                 pod_type=dl.InstanceCatalog.REGULAR_S,
                                                 autoscaler=dl.KubernetesRabbitmqAutoscaler(
                                                     min_replicas=1,
                                                     max_replicas=20,
                                                     queue_length=20
                                                 )
                                             )
                                             )

        """
        return self.project.packages.deploy(package=self,
                                            service_name=service_name,
                                            project_id=self.project_id,
                                            revision=revision,
                                            init_input=init_input,
                                            runtime=runtime,
                                            sdk_version=sdk_version,
                                            agent_versions=agent_versions,
                                            pod_type=pod_type,
                                            bot=bot,
                                            verify=verify,
                                            module_name=module_name,
                                            run_execution_as_process=run_execution_as_process,
                                            execution_timeout=execution_timeout,
                                            drain_time=drain_time,
                                            on_reset=on_reset,
                                            max_attempts=max_attempts,
                                            force=force,
                                            jwt_forward=kwargs.get('jwt_forward', None),
                                            is_global=kwargs.get('is_global', None),
                                            secrets=secrets)

    def checkout(self):
        """
        Checkout as package

        :return:
        """
        return self.packages.checkout(package=self)

    def delete(self) -> bool:
        """
        Delete Package object

        :return: True
        """
        return self.packages.delete(package=self)

    def push(self,
             codebase: Union[entities.GitCodebase, entities.ItemCodebase] = None,
             src_path: str = None,
             package_name: str = None,
             modules: list = None,
             checkout: bool = False,
             revision_increment: str = None,
             service_update: bool = False,
             service_config: dict = None,
             package_type='faas'
             ):
        """
        Push local package

        :param dtlpy.entities.codebase.Codebase codebase: PackageCode object - defines how to store the package code
        :param bool checkout: save package to local checkout
        :param str src_path: location of pacjage codebase folder to zip
        :param str package_name: name of package
        :param list modules: list of PackageModule
        :param str revision_increment: optional - str - version bumping method - major/minor/patch - default = None
        :param  bool service_update: optional - bool - update the service
        :param dict service_config : Service object as dict. Contains the spec of the default service to create.
        :param  str package_type: default is "faas", one of "app", "ml"
        :return: package entity
        :rtype: dtlpy.entities.package.Package
        
        **Example**:

        .. code-block:: python

            package = packages.push(package_name='package_name',
                                    modules=[module],
                                    version='1.0.0',
                                    src_path=os.getcwd())
        """
        return self.project.packages.push(
            package_name=package_name if package_name is not None else self.name,
            modules=modules if modules is not None else self.modules,
            revision_increment=revision_increment,
            codebase=codebase,
            src_path=src_path,
            checkout=checkout,
            service_update=service_update,
            service_config=service_config,
            package_type=package_type
        )

    def pull(self, version=None, local_path=None) -> str:
        """
        Pull local package

        :param str version: version
        :param str local_path: local path

        **Example**:

        .. code-block:: python

            path = package.pull(local_path='local_path')
        """
        return self.packages.pull(package=self,
                                  version=version,
                                  local_path=local_path)

    def build(self, module_name=None, init_inputs=None, local_path=None, from_local=None):
        """
        Instantiate a module from the package code. Returns a loaded instance of the runner class

        :param module_name: Name of the module to build the runner class
        :param str init_inputs: dictionary of the class init variables (if exists). will be used to init the module class
        :param str local_path: local path of the package (if from_local=False - codebase will be downloaded)
        :param bool from_local: bool. if true - codebase will not be downloaded (only use local files)
        :return: dl.BaseServiceRunner
        """
        return self.packages.build(package=self,
                                   module_name=module_name,
                                   local_path=local_path,
                                   init_inputs=init_inputs,
                                   from_local=from_local)

    def open_in_web(self):
        """
        Open the package in web platform

        """
        url = self._client_api._get_resource_url(
            f"projects/{self.project.id}/faas?byCreator=false&byProject=true&byDataloop=false&tab=library&name={self.name}")
        self._client_api._open_in_web(url=url)

    def test(self,
             cwd=None,
             concurrency=None,
             module_name=entities.package_defaults.DEFAULT_PACKAGE_MODULE_NAME,
             function_name=entities.package_defaults.DEFAULT_PACKAGE_FUNCTION_NAME,
             class_name=entities.package_defaults.DEFAULT_PACKAGE_CLASS_NAME,
             entry_point=entities.package_defaults.DEFAULT_PACKAGE_ENTRY_POINT
             ):
        """
        Test local package in local environment.

        :param str cwd: path to the file
        :param int concurrency: the concurrency of the test
        :param str module_name: module name
        :param str function_name: function name
        :param str class_name: class name
        :param str entry_point: the file to run like main.py
        :return: list created by the function that tested the output
        :rtype: list

        **Example**:

        .. code-block:: python

            package.test(cwd='path_to_package',
                        function_name='run')
        """
        return self.project.packages.test_local_package(
            cwd=cwd,
            concurrency=concurrency,
            package=self,
            module_name=module_name,
            function_name=function_name,
            class_name=class_name,
            entry_point=entry_point
        )

    @staticmethod
    def _mockify_input(input_type):
        _json = dict()
        if input_type == 'Dataset':
            _json.update({'dataset_id': 'id'})
        if input_type == 'Item':
            _json.update({'item_id': 'id', 'dataset_id': 'id'})
        if input_type == 'Annotation':
            _json.update({'annotation_id': 'id', 'item_id': 'id', 'dataset_id': 'id'})
        return _json

    def mockify(self, local_path=None, module_name=None, function_name=None):
        if local_path is None:
            local_path = os.getcwd()

        if module_name is None:
            if self.modules:
                module_name = self.modules[0].name
            else:
                raise exceptions.PlatformException('400', 'Package has no modules')

        modules = [module for module in self.modules if module.name == module_name]
        if not modules:
            raise exceptions.PlatformException('404', 'Module not found: {}'.format(module_name))
        module = modules[0]

        if function_name is None:
            funcs = [func for func in module.functions]
            if funcs:
                func = funcs[0]
            else:
                raise exceptions.PlatformException('400', 'Module: {} has no functions'.format(module_name))
        else:
            funcs = [func for func in module.functions if func.name == function_name]
            if not funcs:
                raise exceptions.PlatformException('404', 'Function not found: {}'.format(function_name))
            func = funcs[0]

        mock = dict()
        for module in self.modules:
            mock['module_name'] = module.name
            mock['function_name'] = func.name
            mock['init_params'] = {inpt.name: self._mockify_input(input_type=inpt.type) for inpt in module.init_inputs}
            mock['inputs'] = [{'name': inpt.name, 'value': self._mockify_input(input_type=inpt.type)} for inpt in
                              func.inputs]

        with open(os.path.join(local_path, 'mock.json'), 'w') as f:
            json.dump(mock, f)

    @staticmethod
    def get_ml_metadata(cls=None,
                        available_methods=None,
                        output_type=entities.AnnotationType.CLASSIFICATION,
                        input_type='image',
                        default_configuration: dict = None):
        """
        Create ML metadata for the package
        :param cls: ModelAdapter class, to get the list of available_methods
        :param available_methods: available user function on the adapter.  ['load', 'save', 'predict', 'train']
        :param output_type: annotation type the model create, e.g. dl.AnnotationType.CLASSIFICATION
        :param input_type: input file type the model gets, one of ['image', 'video', 'txt']
        :param default_configuration: default service configuration for the deployed services
        :return:
        """
        user_implemented_methods = ['load', 'save', 'predict', 'train']
        if available_methods is None:
            # default
            available_methods = user_implemented_methods

        if cls is not None:
            # TODO dont check if function is on the adapter - check if the functions is implemented (not raise NotImplemented)
            available_methods = [
                {name: 'BaseModelAdapter' not in getattr(cls, name).__qualname__}
                for name in user_implemented_methods
            ]
        if default_configuration is None:
            default_configuration = dict()
        metadata = {
            'system': {'ml': {'defaultConfiguration': default_configuration,
                              'outputType': output_type,
                              'inputType': input_type,
                              'supportedMethods': available_methods
                              }}}
        return metadata

    class decorators:
        @staticmethod
        def module(name='default-module', description='', init_inputs=None):
            def wrapper(cls: typing.Callable):
                # package_module_dict = package_module.to_json()
                package_module_dict = {"name": name,
                                       "description": description,
                                       "functions": list(),
                                       "className": cls.__name__}
                if init_inputs is not None:
                    package_module_dict.update(initInputs=Package.decorators.parse_io(io_list=init_inputs))
                for member_name, member in inspect.getmembers(cls, predicate=inspect.isfunction):
                    spec = getattr(member, '__dtlpy__', None)
                    if spec is not None:
                        package_module_dict["functions"].append(spec)
                cls.__dtlpy__ = package_module_dict
                return cls

            return wrapper

        @staticmethod
        def function(display_name=None, inputs=None, outputs=None):
            def wrapper(func: typing.Callable):
                if display_name is None:
                    d_name = func.__name__
                else:
                    d_name = display_name
                func.__dtlpy__ = {"name": func.__name__,
                                  "displayName": d_name,
                                  "input": Package.decorators.parse_io(io_list=inputs),
                                  "output": Package.decorators.parse_io(io_list=outputs)}
                return func

            return wrapper

        @staticmethod
        def parse_io(io_list: dict):
            output = list()
            if io_list is not None:
                for io_name, io_type in io_list.items():
                    if isinstance(io_type, Enum):
                        io_type = io_type.name
                    if isinstance(io_type, type):
                        io_type = io_type.__name__
                    else:
                        io_type = str(io_type)
                    output.append({"name": io_name,
                                   "type": str(io_type)})
            return output



"""
Dataloop platform calls
"""
import aiohttp.client_exceptions
import requests_toolbelt
import multiprocessing
import threading
import traceback
import datetime
import requests
import aiohttp
import logging
import asyncio
import certifi
import base64
import enum
import time
import tqdm
import json
import sys
import ssl
import jwt
import os
import io
import concurrent
from concurrent.futures import ThreadPoolExecutor
from requests.adapters import HTTPAdapter
from urllib3.util import Retry
from functools import wraps
import numpy as np
import inspect
from requests.models import Response
from dtlpy.caches.cache import CacheManger, CacheConfig
from .calls_counter import CallsCounter
from .cookie import CookieIO
from .logins import login, logout, login_secret, login_m2m, gate_url_from_host
from .async_utils import AsyncResponse, AsyncUploadStream, AsyncResponseError, AsyncThreadEventLoop
# from .events import Events
from .service_defaults import DEFAULT_ENVIRONMENTS, DEFAULT_ENVIRONMENT
from .aihttp_retry import RetryClient
from .. import miscellaneous, exceptions, __version__

logger = logging.getLogger(name='dtlpy')
threadLock = threading.Lock()


def format_message(message):
    if message and isinstance(message, str):
        return message.replace('\\n', '\n')
    return message


class VerboseLoggingLevel:
    DEBUG = "debug"
    INFO = "info"
    WARNING = "warning"
    ERROR = "error"
    CRITICAL = "critical"


class PlatformError(Exception):
    """
    Error handling for api calls
    """

    def __init__(self, resp):
        msg = ''
        if hasattr(resp, 'status_code'):
            msg += '<Response [{}]>'.format(resp.status_code)
        if hasattr(resp, 'reason'):
            msg += '<Reason [{}]>'.format(format_message(resp.reason))
        elif hasattr(resp, 'text'):
            msg += '<Reason [{}]>'.format(format_message(resp.text))
        super().__init__(msg)


class Callbacks:
    def __init__(self):
        self._callbacks = {}

    class CallbackEvent(str, enum.Enum):
        DATASET_EXPORT = 'datasetExport'
        ITEMS_UPLOAD = 'itemUpload'

    def add(self, event, func):

        if not callable(func):
            raise ValueError(f"The provided callback for {event} is not callable")
        if event not in list(self.CallbackEvent):
            raise ValueError(f"Unknown event: {event!r}, allowed events are: {list(self.CallbackEvent)}")
        self._callbacks[event] = func

    def get(self, name):
        return self._callbacks.get(name)

    def run_on_event(self, event, context, progress):
        callback = self.get(event)
        if callback is not None:
            callback(progress=progress, context=context)


class Verbose:
    __DEFAULT_LOGGING_LEVEL = 'warning'
    __DEFAULT_DISABLE_PROGRESS_BAR = False
    __DEFAULT_PRINT_ALL_RESPONSES = False
    __PRINT_ERROR_LOGS = False
    __DEFAULT_PROGRESS_BAR_SETTINGS = {
        'Iterate Pages': False,
        'Command Progress': False,
        'Download Dataset': False,
        'Download Item': False,
        'Upload Items': False,
        'Download Annotations': False,
        'Upload Annotations': False,
        'Convert Annotations': False
    }

    def __init__(self, cookie):
        self.cookie = cookie
        dictionary = self.cookie.get('verbose')
        if isinstance(dictionary, dict):
            self.from_cookie(dictionary)
        else:
            self._logging_level = self.__DEFAULT_LOGGING_LEVEL
            self._disable_progress_bar = self.__DEFAULT_DISABLE_PROGRESS_BAR
            self._print_all_responses = self.__DEFAULT_PRINT_ALL_RESPONSES
            self._print_error_logs = self.__PRINT_ERROR_LOGS
            self._progress_bar_settings = self.__DEFAULT_PROGRESS_BAR_SETTINGS
            if os.getenv('DTLPY_REFRESH_TOKEN_METHOD', "") == "proxy":
                self._print_error_logs = True
            self.to_cookie()

    def to_cookie(self):
        dictionary = {'logging_level': self._logging_level,
                      'disable_progress_bar': self._disable_progress_bar,
                      'print_all_responses': self._print_all_responses,
                      'print_error_logs': self._print_error_logs,
                      'progress_bar_setting': json.dumps(self._progress_bar_settings)
                      }
        self.cookie.put(key='verbose', value=dictionary)

    def from_cookie(self, dictionary):
        self._logging_level = dictionary.get('logging_level', self.__DEFAULT_LOGGING_LEVEL)
        self._disable_progress_bar = dictionary.get('disable_progress_bar', self.__DEFAULT_DISABLE_PROGRESS_BAR)
        self._print_all_responses = dictionary.get('print_all_responses', self.__DEFAULT_PRINT_ALL_RESPONSES)
        self._print_error_logs = dictionary.get('print_error_logs', self.__PRINT_ERROR_LOGS)
        progress_bar_settings = dictionary.get('progress_bar_setting', None)
        if progress_bar_settings is None:
            self._progress_bar_settings = self.__DEFAULT_PROGRESS_BAR_SETTINGS
        else:
            self._progress_bar_settings = json.loads(progress_bar_settings)

    @property
    def disable_progress_bar_iterate_pages(self):
        return self._disable_progress_bar or self._progress_bar_settings.get('Iterate Pages', False)

    @disable_progress_bar_iterate_pages.setter
    def disable_progress_bar_iterate_pages(self, val):
        self._progress_bar_settings['Iterate Pages'] = val
        self.to_cookie()

    @property
    def disable_progress_bar_command_progress(self):
        return self._disable_progress_bar or self._progress_bar_settings.get('Command Progress', False)

    @disable_progress_bar_command_progress.setter
    def disable_progress_bar_command_progress(self, val):
        self._progress_bar_settings['Command Progress'] = val
        self.to_cookie()

    @property
    def disable_progress_bar_download_item(self):
        return self._disable_progress_bar or self._progress_bar_settings.get('Download Item', False)

    @disable_progress_bar_download_item.setter
    def disable_progress_bar_download_item(self, val):
        self._progress_bar_settings['Download Item'] = val
        self.to_cookie()

    @property
    def disable_progress_bar_download_dataset(self):
        return self._disable_progress_bar or self._progress_bar_settings.get('Download Dataset', False)

    @disable_progress_bar_download_dataset.setter
    def disable_progress_bar_download_dataset(self, val):
        self._progress_bar_settings['Download Dataset'] = val
        self.to_cookie()

    @property
    def disable_progress_bar_upload_items(self):
        return self._disable_progress_bar or self._progress_bar_settings.get('Upload Items', False)

    @disable_progress_bar_upload_items.setter
    def disable_progress_bar_upload_items(self, val):
        self._progress_bar_settings['Upload Items'] = val
        self.to_cookie()

    @property
    def disable_progress_bar_download_annotations(self):
        return self._disable_progress_bar or self._progress_bar_settings.get('Download Annotations', False)

    @disable_progress_bar_download_annotations.setter
    def disable_progress_bar_download_annotations(self, val):
        self._progress_bar_settings['Download Annotations'] = val
        self.to_cookie()

    @property
    def disable_progress_bar_upload_annotations(self):
        return self._disable_progress_bar or self._progress_bar_settings.get('Upload Annotations', False)

    @disable_progress_bar_upload_annotations.setter
    def disable_progress_bar_upload_annotations(self, val):
        self._progress_bar_settings['Upload Annotations'] = val
        self.to_cookie()

    @property
    def disable_progress_bar_convert_annotations(self):
        return self._disable_progress_bar or self._progress_bar_settings.get('Convert Annotations', False)

    @disable_progress_bar_convert_annotations.setter
    def disable_progress_bar_convert_annotations(self, val):
        self._progress_bar_settings['Convert Annotations'] = val
        self.to_cookie()

    @property
    def disable_progress_bar(self):
        return self._disable_progress_bar

    @disable_progress_bar.setter
    def disable_progress_bar(self, val):
        self._disable_progress_bar = val
        self.to_cookie()

    @property
    def logging_level(self):
        return self._logging_level

    @logging_level.setter
    def logging_level(self, val):
        self._logging_level = val
        # set log level
        logging.getLogger(name='dtlpy').handlers[0].setLevel(logging._nameToLevel[self._logging_level.upper()])
        # write to cookie
        self.to_cookie()

    @property
    def print_all_responses(self):
        return self._print_all_responses

    @print_all_responses.setter
    def print_all_responses(self, val):
        self._print_all_responses = val
        self.to_cookie()

    @property
    def print_error_logs(self):
        return self._print_error_logs

    @print_error_logs.setter
    def print_error_logs(self, val):
        self._print_error_logs = val
        self.to_cookie()


class CacheMode:
    __DEFAULT_ENABLE_CACHE = True
    __DEFAULT_CHUNK_CACHE = 200000

    def __init__(self, cookie):
        self.cookie = cookie
        dictionary = self.cookie.get('cache_mode')
        if isinstance(dictionary, dict):
            self.from_cookie(dictionary)
        else:
            self._enable_cache = self.__DEFAULT_ENABLE_CACHE
            self._chunk_cache = self.__DEFAULT_CHUNK_CACHE
            self.to_cookie()

    def to_cookie(self):
        dictionary = {'enable_cache': self._enable_cache,
                      'chunk_cache': self._chunk_cache}
        self.cookie.put(key='cache_mode', value=dictionary)

    def from_cookie(self, dictionary):
        self._enable_cache = dictionary.get('enable_cache', self.__DEFAULT_ENABLE_CACHE)
        self._chunk_cache = dictionary.get('chunk_cache', self.__DEFAULT_CHUNK_CACHE)

    @property
    def enable_cache(self):
        return self._enable_cache

    @enable_cache.setter
    def enable_cache(self, val: bool):
        if not isinstance(val, bool):
            raise exceptions.PlatformException(error=400,
                                               message="input must be of type bool")
        self._enable_cache = val
        self.to_cookie()

    @property
    def chunk_cache(self):
        return self._chunk_cache

    @chunk_cache.setter
    def chunk_cache(self, val):
        self._chunk_cache = val
        self.to_cookie()


class SDKCache:
    __DEFAULT_USE_CACHE = False
    __DEFAULT_CACHE_PATH = os.path.join(os.path.expanduser('~'), '.dataloop', 'obj_cache')
    __DEFAULT_CACHE_PATH_BIN = os.path.join(os.path.expanduser('~'), '.dataloop')
    __DEFAULT_CONFIGS_CACHE = CacheConfig().to_string()
    __DEFAULT_BINARY_CACHE_SIZE = 1000

    def __init__(self, cookie):
        self.cookie = cookie
        dictionary = self.cookie.get('cache_configs')
        if isinstance(dictionary, dict):
            self.from_cookie(dictionary)
        else:
            self._cache_path = self.__DEFAULT_CACHE_PATH
            self._cache_path_bin = self.__DEFAULT_CACHE_PATH_BIN
            self._configs = self.__DEFAULT_CONFIGS_CACHE
            self._bin_size = self.__DEFAULT_BINARY_CACHE_SIZE
            self._use_cache = self.__DEFAULT_USE_CACHE
            self.to_cookie()

    def to_cookie(self):
        dictionary = {'cache_path': self._cache_path,
                      'cache_path_bin': self._cache_path_bin,
                      'configs': self._configs,
                      'bin_size': self._bin_size,
                      'use_cache': self._use_cache}
        self.cookie.put(key='cache_configs', value=dictionary)

    def from_cookie(self, dictionary):
        self._cache_path = dictionary.get('cache_path', self.__DEFAULT_CACHE_PATH)
        self._cache_path_bin = dictionary.get('cache_path_bin', self.__DEFAULT_CACHE_PATH_BIN)
        self._configs = dictionary.get('configs', self.__DEFAULT_CONFIGS_CACHE)
        self._bin_size = dictionary.get('bin_size', self.__DEFAULT_BINARY_CACHE_SIZE)
        self._use_cache = dictionary.get('use_cache', self.__DEFAULT_USE_CACHE)

    @property
    def cache_path(self):
        return self._cache_path

    @property
    def cache_path_bin(self):
        return self._cache_path_bin

    @cache_path_bin.setter
    def cache_path_bin(self, val: str):
        if not isinstance(val, str):
            raise exceptions.PlatformException(error=400,
                                               message="input must be of type str")
        self._cache_path_bin = val
        os.environ['DEFAULT_CACHE_PATH'] = val
        self.to_cookie()

    @property
    def use_cache(self):
        return self._use_cache

    @use_cache.setter
    def use_cache(self, val: bool):
        if not isinstance(val, bool):
            raise exceptions.PlatformException(error=400,
                                               message="input must be of type bool")
        self._use_cache = val
        self.to_cookie()

    @property
    def configs(self):
        return self._configs

    @configs.setter
    def configs(self, val):
        if isinstance(val, CacheConfig):
            val = val.to_string()
        if not isinstance(val, str):
            raise exceptions.PlatformException(error=400,
                                               message="input must be of type str or CacheConfig")
        self._configs = val
        self.to_cookie()

    @property
    def bin_size(self):
        return self._bin_size

    @bin_size.setter
    def bin_size(self, val: int):
        if not isinstance(val, int):
            raise exceptions.PlatformException(error=400,
                                               message="input must be of type int")
        self._bin_size = val
        self.to_cookie()


class Attributes2:
    __DEFAULT_USE_ATTRIBUTE = False

    def __init__(self, cookie):
        self.cookie = cookie
        dictionary = self.cookie.get('use_attributes_2')
        if isinstance(dictionary, dict):
            self.from_cookie(dictionary)
        else:
            self._use_attributes_2 = self.__DEFAULT_USE_ATTRIBUTE
            self.to_cookie()

    def to_cookie(self):
        dictionary = {'use_attributes_2': self._use_attributes_2}
        self.cookie.put(key='use_attributes_2', value=dictionary)

    def from_cookie(self, dictionary):
        self._use_attributes_2 = dictionary.get('use_attributes_2', self.__DEFAULT_USE_ATTRIBUTE)

    @property
    def use_attributes_2(self):
        return self._use_attributes_2

    @use_attributes_2.setter
    def use_attributes_2(self, val: bool):
        if not isinstance(val, bool):
            raise exceptions.PlatformException(error=400,
                                               message="input must be of type bool")
        self._use_attributes_2 = val
        os.environ["USE_ATTRIBUTE_2"] = json.dumps(val)
        self.to_cookie()


class Decorators:
    @staticmethod
    def token_expired_decorator(method):
        @wraps(method)
        def decorated_method(inst, *args, **kwargs):
            # save event
            frm = inspect.stack()[1]

            # before the method call
            kwargs.update({'stack': frm})
            if inst.token_expired():
                if inst.renew_token_method() is False:
                    raise exceptions.PlatformException('600', 'Token expired, Please login.'
                                                              '\nSDK login options: dl.login(), dl.login_token(), '
                                                              'dl.login_m2m()'
                                                              '\nCLI login options: dlp login, dlp login-token, '
                                                              'dlp login-m2m')
            # the actual method call
            result = method(inst, *args, **kwargs)
            # after the method call
            return result

        return decorated_method


class ApiClient:
    """
    API calls to Dataloop gate
    """

    def __init__(self, token=None, num_processes=None, cookie_filepath=None):
        ############
        # Initiate #
        ############
        # define local params - read only once from cookie file
        self.lock = threading.Lock()
        self.renew_token_method = self.renew_token
        self.is_cli = False
        self.session = None
        self.default_headers = dict()
        self._token = None
        self._environments = None
        self._environment = None
        self._verbose = None
        self._callbacks = None
        self._cache_state = None
        self._attributes_mode = None
        self._cache_configs = None
        self._sdk_cache = None
        self._fetch_entities = None
        # define other params
        self.last_response = None
        self.last_request = None
        self.platform_exception = None
        self.last_curl = None
        self.minimal_print = True
        # start refresh token
        self.refresh_token_active = True
        # event and pools
        self._thread_pools = dict()
        self._event_loop = None
        self._login_domain = None
        self.__gate_url_for_requests = None

        # TODO- remove before release - only for debugging
        self._stopped_pools = list()

        if cookie_filepath is None:
            self.cookie_io = CookieIO.init()
        else:
            self.cookie_io = CookieIO(path=cookie_filepath)
        assert isinstance(self.cookie_io, CookieIO)
        self.state_io = CookieIO.init_local_cookie(create=False)
        assert isinstance(self.state_io, CookieIO)

        ##################
        # configurations #
        ##################
        # check for proxies in connection
        self.check_proxy()

        # set token if input
        if token is not None:
            self.token = token

        # STDOUT
        self.remove_keys_list = ['contributors', 'url', 'annotations', 'items', 'export', 'directoryTree',
                                 'attributes', 'partitions', 'metadata', 'stream', 'createdAt', 'updatedAt', 'arch']

        # API calls counter
        counter_filepath = os.path.join(os.path.dirname(self.cookie_io.COOKIE), 'calls_counter.json')
        self.calls_counter = CallsCounter(filepath=counter_filepath)

        # create a global thread pool to run multi threading
        if num_processes is None:
            num_processes = 3 * multiprocessing.cpu_count()
        self._num_processes = num_processes
        self._thread_pools_names = {'item.download': num_processes,
                                    'item.status_update': num_processes,
                                    'item.page': num_processes,
                                    'annotation.upload': num_processes,
                                    'annotation.download': num_processes,
                                    'annotation.update': num_processes,
                                    'entity.create': num_processes,
                                    'dataset.download': num_processes}
        # set logging level
        logging.getLogger(name='dtlpy').handlers[0].setLevel(logging._nameToLevel[self.verbose.logging_level.upper()])
        os.environ["USE_ATTRIBUTE_2"] = json.dumps(self.attributes_mode.use_attributes_2)

        self.cache = None
        #######################
        # start event tracker #
        # self.event_tracker = Events(client_api=self)
        # self.event_tracker.daemon = True
        # self.event_tracker.start()
        self.upload_session_timeout = int(os.environ.get('UPLOAD_SESSION_TIMEOUT', 0))
        self.upload_chunk_timeout = int(os.environ.get('UPLOAD_CHUNK_TIMEOUT', 2 * 60))

    @property
    def event_loop(self):
        self.lock.acquire()
        if self._event_loop is None:
            self._event_loop = self.create_event_loop_thread()
        elif not self._event_loop.loop.is_running():
            if self._event_loop.is_alive():
                self._event_loop.stop()
            self._event_loop = self.create_event_loop_thread()
        self.lock.release()
        return self._event_loop

    def build_cache(self, cache_config=None):
        if cache_config is None:
            cache_config_json = os.environ.get('CACHE_CONFIG', None)
            if cache_config_json is None:
                if self.sdk_cache.use_cache:
                    cache_config = CacheConfig.from_string(cls=CacheConfig, base64_string=self.sdk_cache.configs)
            else:
                cache_config = CacheConfig.from_string(cls=CacheConfig, base64_string=cache_config_json)
        if cache_config:
            # cache paths
            if os.environ.get('DEFAULT_CACHE_PATH', None) is None:
                os.environ['DEFAULT_CACHE_PATH'] = self.sdk_cache.cache_path_bin
            else:
                self.sdk_cache.cache_path_bin = os.environ['DEFAULT_CACHE_PATH']

            if not os.path.isdir(self.sdk_cache.cache_path_bin):
                os.makedirs(self.sdk_cache.cache_path_bin, exist_ok=True)

            if not os.path.isfile(os.path.join(self.sdk_cache.cache_path_bin, 'cacheConfig.json')):
                os.makedirs(self.sdk_cache.cache_path_bin, exist_ok=True)

            if isinstance(cache_config, str):
                self.sdk_cache.configs = cache_config
                cache_config = CacheConfig.from_string(cls=CacheConfig, base64_string=cache_config)
            elif isinstance(cache_config, CacheConfig):
                self.sdk_cache.configs = cache_config.to_string()
            else:
                raise Exception("config should be of type str or CacheConfig")
            try:
                self.cache = CacheManger(cache_configs=[cache_config], bin_cache_size=self.sdk_cache.bin_size)
                self.cache.ping()
                self.sdk_cache.use_cache = True
            except Exception as e:
                logger.warning("Cache build error {}".format(e))
                self.cache = None

    def __del__(self):
        for name, pool in self._thread_pools.items():
            pool.shutdown()
        self.event_loop.stop()

    def _build_request_headers(self, headers=None):
        if headers is None:
            headers = dict()
        if not isinstance(headers, dict):
            raise exceptions.PlatformException(
                error='400',
                message="Input 'headers' must be a dictionary, got: {}".format(type(headers)))
        headers.update(self.default_headers)
        headers.update(self.auth)
        headers.update({'User-Agent': requests_toolbelt.user_agent('dtlpy', __version__)})
        return headers

    @property
    def num_processes(self):
        return self._num_processes

    @num_processes.setter
    def num_processes(self, num_processes):
        if num_processes == self._num_processes:
            # same number. no need to do anything
            return
        self._num_processes = num_processes
        for pool_name in self._thread_pools_names:
            self._thread_pools_names[pool_name] = num_processes

        for pool in self._thread_pools:
            self._thread_pools[pool].shutdown()
        self._thread_pools = dict()

    def create_event_loop_thread(self):
        loop = asyncio.new_event_loop()
        event_loop = AsyncThreadEventLoop(loop=loop,
                                          n=self._num_processes)
        event_loop.daemon = True
        event_loop.start()
        time.sleep(1)
        return event_loop

    def thread_pools(self, pool_name):
        if pool_name not in self._thread_pools_names:
            raise ValueError('unknown thread pool name: {}. known name: {}'.format(
                pool_name,
                list(self._thread_pools_names.keys())))
        num_processes = self._thread_pools_names[pool_name]
        if pool_name not in self._thread_pools or self._thread_pools[pool_name]._shutdown:
            self._thread_pools[pool_name] = ThreadPoolExecutor(max_workers=num_processes)
        pool = self._thread_pools[pool_name]
        assert isinstance(pool, concurrent.futures.ThreadPoolExecutor)
        return pool

    @property
    def verify(self):
        environments = self.environments
        verify = True
        if self.environment in environments:
            if 'verify_ssl' in environments[self.environment]:
                verify = environments[self.environment]['verify_ssl']
        return verify

    @property
    def use_ssl_context(self):
        environments = self.environments
        use_ssl_context = False
        if self.environment in environments:
            if 'use_ssl_context' in environments[self.environment]:
                use_ssl_context = environments[self.environment]['use_ssl_context']
        return use_ssl_context

    @property
    def auth(self):
        return {'authorization': 'Bearer ' + self.token}

    @property
    def environment(self):
        _environment = self._environment
        if _environment is None:
            _environment = self.cookie_io.get('url')
            if _environment is None:
                _environment = DEFAULT_ENVIRONMENT
            self._environment = _environment
        return _environment

    @environment.setter
    def environment(self, env):
        self._environment = env
        self.cookie_io.put('url', env)

    @property
    def fetch_entities(self):
        if self._fetch_entities is None:
            self._fetch_entities = self.cookie_io.get('fetch_entities')
            if self._fetch_entities is None:
                self.fetch_entities = True  # default
        return self._fetch_entities

    @fetch_entities.setter
    def fetch_entities(self, val):
        self._fetch_entities = val
        self.cookie_io.put('fetch_entities', val)

    @property
    def environments(self):
        """
        List of known environments
        :return:
        """
        # get environment login parameters
        _environments = self._environments
        if _environments is None:
            # take from cookie
            _environments = self.cookie_io.get('login_parameters')
            # if cookie is None  - init with defaults
            if _environments is None:
                # default
                _environments = DEFAULT_ENVIRONMENTS
                # save to local variable
                self.environments = _environments
            else:
                # save from cookie to ram
                self._environments = _environments
        return _environments

    @environments.setter
    def environments(self, env_dict):
        self._environments = env_dict
        self.cookie_io.put(key='login_parameters', value=self._environments)

    @property
    def verbose(self):
        if self._verbose is None:
            self._verbose = Verbose(cookie=self.cookie_io)
        assert isinstance(self._verbose, Verbose)
        return self._verbose

    @property
    def cache_state(self):
        if self._cache_state is None:
            self._cache_state = CacheMode(cookie=self.cookie_io)
        assert isinstance(self._cache_state, CacheMode)
        return self._cache_state

    @property
    def attributes_mode(self):
        if self._attributes_mode is None:
            self._attributes_mode = Attributes2(cookie=self.cookie_io)
        assert isinstance(self._attributes_mode, Attributes2)
        return self._attributes_mode

    @property
    def sdk_cache(self):
        if self._sdk_cache is None:
            self._sdk_cache = SDKCache(cookie=self.cookie_io)
        assert isinstance(self._sdk_cache, SDKCache)
        return self._sdk_cache

    @property
    def callbacks(self):
        if self._callbacks is None:
            self._callbacks = Callbacks()
        assert isinstance(self._callbacks, Callbacks)
        return self._callbacks

    def add_callback(self, event, func):
        """
        function to add callback to the client
        :param event: dl.CallbackEvent enum, name of the callback
        :param func: function to call with 2 arguments: progress and context
        """
        self.callbacks.add(event, func)

    @property
    def token(self):
        _token = self._token
        if _token is None:
            environments = self.environments
            if self.environment in environments:
                if 'token' in environments[self.environment]:
                    _token = environments[self.environment]['token']
        return _token

    @token.setter
    def token(self, token):
        # set to variable
        self._token = token
        self.refresh_token = None
        # set to cookie file
        environments = self.environments
        if self.environment in environments:
            environments[self.environment]['token'] = token
        else:
            environments[self.environment] = {'token': token}
        self.environments = environments

    @property
    def refresh_token(self):
        environments = self.environments
        refresh_token = None
        if self.environment in environments:
            if 'refresh_token' in environments[self.environment]:
                refresh_token = environments[self.environment]['refresh_token']
        return refresh_token

    @refresh_token.setter
    def refresh_token(self, token):
        environments = self.environments
        if self.environment in environments:
            environments[self.environment]['refresh_token'] = token
        else:
            environments[self.environment] = {'refresh_token': token}
        self.refresh_token_active = True
        self.environments = environments

    def add_environment(self, environment,
                        audience=None,
                        client_id=None,
                        auth0_url=None,
                        verify_ssl=True,
                        token=None,
                        refresh_token=None,
                        alias=None,
                        use_ssl_context=False,
                        gate_url=None,
                        url=None,
                        login_domain=None
                        ):
        environments = self.environments
        if environment in environments:
            logger.warning('Environment exists. Overwriting. env: {}'.format(environment))
        if alias is not None:
            keys_to_remove = [e for e, v in environments.items() if v.get('alias') == alias]
            for e in keys_to_remove:
                logger.warning('Alias exists. Overwriting. alias: {}'.format(alias))
                environments.pop(e)
        environments[environment] = {'audience': audience,
                                     'client_id': client_id,
                                     'auth0_url': auth0_url,
                                     'alias': alias,
                                     'token': token,
                                     'gate_url': gate_url,
                                     'refresh_token': refresh_token,
                                     'verify_ssl': verify_ssl,
                                     'use_ssl_context': use_ssl_context,
                                     'url': url,
                                     'login_domain': login_domain}
        self.environments = environments

    def info(self, with_token=True):
        """
        Return a dictionary with current information: env, user, token
        :param with_token:
        :return:
        """
        user_email = 'null'
        if self.token is not None:
            # oxsec-disable jwt-signature-disabled - Client-side SDK: signature verification disabled intentionally to extract user email; server validates on API calls
            payload = jwt.decode(
                self.token,
                options={
                    "verify_signature": False,
                    "verify_exp": False,
                    "verify_aud": False,
                    "verify_iss": False,
                }
            )
            user_email = payload['email']
        information = {'environment': self.environment,
                       'user_email': user_email}
        if with_token:
            information['token'] = self.token
        return information

    @property
    def base_gate_url(self):
        if self.__gate_url_for_requests is None:
            self.__gate_url_for_requests = self.environment
            internal_requests_url = os.environ.get('INTERNAL_REQUESTS_URL', None)
            if internal_requests_url is not None:
                self.__gate_url_for_requests = internal_requests_url
        return self.__gate_url_for_requests

    def export_curl_request(self, req_type, path, headers=None, json_req=None, files=None, data=None):
        curl, prepared = self._build_gen_request(req_type=req_type,
                                                 path=path,
                                                 headers=headers,
                                                 json_req=json_req,
                                                 files=files,
                                                 data=data)
        return curl

    def _build_gen_request(self, req_type, path, headers, json_req, files, data):
        req_type = req_type.upper()
        valid_request_type = ['GET', 'DELETE', 'POST', 'PUT', 'PATCH']
        assert req_type in valid_request_type, '[ERROR] type: %s NOT in valid requests' % req_type

        # prepare request
        req = requests.Request(method=req_type,
                               url=self.base_gate_url + path,
                               json=json_req,
                               files=files,
                               data=data,
                               headers=self._build_request_headers(headers=headers))
        # prepare to send
        prepared = req.prepare()
        # save curl for debug
        command = "curl -X {method} -H {headers} -d '{data}' '{uri}'"
        method = prepared.method
        uri = prepared.url
        data = prepared.body
        headers = ['"{0}: {1}"'.format(k, v) for k, v in prepared.headers.items()]
        headers = " -H ".join(headers)
        curl = command.format(method=method, headers=headers, data=data, uri=uri)
        return curl, prepared

    def _convert_json_to_response(self, response_json):
        the_response = Response()
        the_response._content = json.dumps(response_json).encode('utf-8')
        return the_response

    def _cache_on(self, request):
        if self.cache is not None and self.sdk_cache.use_cache:
            pure_request = request.split('?')[0]
            valid_req = ['annotation', 'item', 'dataset', 'project', 'task', 'assignment']
            for req_type in valid_req:
                if req_type in pure_request:
                    return True
        return False

    @Decorators.token_expired_decorator
    def gen_request(self, req_type, path, data=None, json_req=None, files=None, stream=False, headers=None,
                    log_error=True, dataset_id=None, **kwargs):
        """
        Generic request from platform
        :param req_type: type of the request: GET, POST etc
        :param path: url (without host header - take from environment)
        :param data: data to pass to request
        :param json_req: json to pass to request
        :param files: files to pass to request
        :param stream: stream to pass the request
        :param headers: headers to pass to request. auth will be added to it
        :param log_error: if true - print the error log of the request
        :param dataset_id: dataset id needed in stream True
        :param kwargs: kwargs
        :return:
        """
        success, resp, cache_values = False, None, []
        if self.cache is None and 'sdk' not in path:
            self.build_cache()
        if req_type.lower() not in ['patch', 'put', 'post', 'delete'] and self._cache_on(request=path):
            try:
                if stream:
                    if dataset_id is None:
                        raise ValueError("must provide a dataset id")
                    success, cache_values = self.cache.read_stream(request_path=path, dataset_id=dataset_id)

                else:
                    success, cache_values = self.cache.read(request_path=path)
                if success:
                    resp = self._convert_json_to_response(cache_values)
            except Exception as e:
                logger.warning("Cache error {}".format(e))
                success, resp = False, None

        if not success and not resp:
            success, resp = self._gen_request(req_type=req_type,
                                              path=path,
                                              data=data,
                                              json_req=json_req,
                                              files=files,
                                              stream=stream,
                                              headers=headers,
                                              log_error=log_error)

            if success and self._cache_on(request=path):
                try:
                    if stream:
                        res = self.cache.write_stream(request_path=path,
                                                      response=resp,
                                                      dataset_id=dataset_id)
                        if res != '':
                            resp = self._convert_json_to_response(res)
                    else:
                        if req_type == 'delete':
                            self.cache.invalidate(path=path)
                        else:
                            try:
                                resp_list = resp.json()
                                write = True
                                if isinstance(resp_list, list):
                                    pass
                                elif isinstance(resp_list, dict):
                                    if 'hasNextPage' in resp_list:
                                        resp_list = resp_list['items']
                                    elif 'id' in resp_list:
                                        resp_list = [resp_list]
                                    else:
                                        write = False
                                else:
                                    raise exceptions.PlatformException(error='400', message="unsupported return type")
                                if write:
                                    self.cache.write(list_entities_json=resp_list)
                            except:
                                raise exceptions.PlatformException(error='400', message="failed to set cache")
                except Exception as e:
                    logger.warning("Cache error {}".format(e))
                    self.cache = None
        # only for projects events
        # if success:
        #     if 'stack' in kwargs:
        #         self.event_tracker.put(event=kwargs.get('stack'), resp=resp, path=path)
        return success, resp

    def _gen_request(self, req_type, path, data=None, json_req=None, files=None, stream=False, headers=None,
                     log_error=True):
        """
        Generic request from platform
        :param req_type: type of the request: GET, POST etc
        :param path: url (without host header - take from environment)
        :param data: data to pass to request
        :param json_req: json to pass to request
        :param files: files to pass to request
        :param stream: stream to pass the request
        :param headers: headers to pass to request. auth will be added to it
        :param log_error: if true - print the error log of the request
        :return:
        """
        curl, prepared = self._build_gen_request(req_type=req_type,
                                                 path=path,
                                                 headers=headers,
                                                 json_req=json_req,
                                                 files=files,
                                                 data=data)
        self.last_curl = curl
        self.last_request = prepared
        # send request
        try:
            resp = self.send_session(prepared=prepared, stream=stream)
        except Exception:
            logger.error(self.print_request(req=prepared, to_return=True))
            raise
        self.last_response = resp
        # handle output
        if not resp.ok:
            self.print_bad_response(resp, log_error=log_error and not self.is_cli)
            return_type = False
        else:
            try:
                # print only what is printable (dont print get steam etc..)
                if not stream:
                    self.print_response(resp)
            except ValueError:
                # no JSON returned
                pass
            return_type = True
        return return_type, resp

    @Decorators.token_expired_decorator
    async def gen_async_request(self,
                                req_type,
                                path,
                                data=None,
                                json_req=None,
                                files=None,
                                stream=None,
                                headers=None,
                                log_error=True,
                                filepath=None,
                                chunk_size=8192,
                                pbar=None,
                                is_dataloop=True,
                                **kwargs):
        req_type = req_type.upper()
        valid_request_type = ['GET', 'DELETE', 'POST', 'PUT', 'PATCH']
        assert req_type in valid_request_type, '[ERROR] type: %s NOT in valid requests' % req_type

        # prepare request
        if is_dataloop:
            full_url = self.base_gate_url + path
            headers_req = self._build_request_headers(headers=headers)
        else:
            full_url = path
            headers = dict()
            headers_req = headers

        if headers is not None:
            if not isinstance(headers, dict):
                raise exceptions.PlatformException(error='400', message="Input 'headers' must be a dictionary")
            for k, v in headers.items():
                headers_req[k] = v
        req = requests.Request(method=req_type,
                               url=full_url,
                               json=json_req,
                               files=files,
                               data=data,
                               headers=headers_req)
        # prepare to send
        prepared = req.prepare()
        # save curl for debug
        command = "curl -X {method} -H {headers} -d '{data}' '{uri}'"
        headers = ['"{0}: {1}"'.format(k, v) for k, v in prepared.headers.items()]
        headers = " -H ".join(headers)
        curl = command.format(method=prepared.method,
                              headers=headers,
                              data=prepared.body,
                              uri=prepared.url)
        self.last_curl = curl
        self.last_request = prepared
        # send request
        try:
            timeout = aiohttp.ClientTimeout(total=0)
            async with RetryClient(headers=headers_req,
                                   timeout=timeout) as session:
                try:
                    async with session._request(request=session._client.request,
                                                url=self.base_gate_url + path,
                                                method=req_type,
                                                json=json_req,
                                                data=data,
                                                headers=headers_req,
                                                chunked=stream,
                                                retry_attempts=5,
                                                ssl=self.verify,
                                                retry_exceptions={aiohttp.client_exceptions.ClientOSError,
                                                                  aiohttp.client_exceptions.ServerDisconnectedError,
                                                                  aiohttp.client_exceptions.ClientPayloadError},
                                                raise_for_status=False) as request:
                        if stream:
                            pbar = self.__get_pbar(pbar=pbar,
                                                   total_length=request.headers.get("content-length"))
                            if filepath is not None:
                                to_close = False
                                if isinstance(filepath, str):
                                    to_close = True
                                    buffer = open(filepath, 'wb')
                                elif isinstance(filepath, io.BytesIO):
                                    pass
                                else:
                                    raise ValueError('unknown data type to write file: {}'.format(type(filepath)))
                                try:
                                    while True:
                                        chunk = await request.content.read(chunk_size)
                                        await asyncio.sleep(0)
                                        if not chunk:
                                            break
                                        buffer.write(chunk)
                                        if pbar is not None:
                                            pbar.update(len(chunk))
                                finally:
                                    if to_close:
                                        buffer.close()

                            if pbar is not None:
                                pbar.close()
                        text = await request.text()
                        try:
                            _json = await request.json()
                        except Exception:
                            _json = dict()
                        response = AsyncResponse(text=text,
                                                 _json=_json,
                                                 async_resp=request)
                except Exception as err:
                    response = AsyncResponseError(error=err, trace=traceback.format_exc())
                finally:
                    with threadLock:
                        self.calls_counter.add()
        except Exception:
            logger.error(self.print_request(req=prepared, to_return=True))
            raise
        self.last_response = response
        # handle output
        if not response.ok:
            self.print_bad_response(response, log_error=log_error and not self.is_cli)
            return_type = False
        else:
            try:
                # print only what is printable (dont print get steam etc..)
                if not stream:
                    self.print_response(response)
            except ValueError:
                # no JSON returned
                pass
            return_type = True
        return return_type, response

    @Decorators.token_expired_decorator
    async def upload_file_async(self,
                                to_upload,
                                item_type,
                                item_size,
                                remote_url,
                                uploaded_filename,
                                remote_path=None,
                                callback=None,
                                mode='skip',
                                item_metadata=None,
                                headers=None,
                                item_description=None,
                                **kwargs):
        headers = self._build_request_headers(headers=headers)
        pbar = None
        if callback is None:
            if item_size > 10e6:
                # size larger than 10MB
                pbar = tqdm.tqdm(total=item_size,
                                 unit="B",
                                 unit_scale=True,
                                 unit_divisor=1024,
                                 position=1,
                                 file=sys.stdout,
                                 disable=self.verbose.disable_progress_bar_upload_items,
                                 desc='Upload Items')

                def callback(bytes_read):
                    pbar.update(bytes_read)
            else:
                def callback(bytes_read):
                    pass

        timeout = aiohttp.ClientTimeout(total=self.upload_session_timeout)
        async with aiohttp.ClientSession(headers=headers, timeout=timeout) as session:
            try:
                form = aiohttp.FormData({})
                form.add_field('type', item_type)
                form.add_field('path', os.path.join(remote_path, uploaded_filename).replace('\\', '/'))
                if item_metadata is not None:
                    form.add_field('metadata', json.dumps(item_metadata))
                if item_description is not None:
                    form.add_field('description', item_description)
                form.add_field('file', AsyncUploadStream(buffer=to_upload,
                                                         callback=callback,
                                                         name=uploaded_filename,
                                                         chunk_timeout=self.upload_chunk_timeout))
                url = '{}?mode={}'.format(self.base_gate_url + remote_url, mode)

                # use SSL context
                ssl_context = None
                if self.use_ssl_context:
                    ssl_context = ssl.create_default_context(cafile=certifi.where())
                async with session.post(url,
                                        data=form,
                                        verify_ssl=self.verify,
                                        ssl=ssl_context) as resp:
                    self.last_request = resp.request_info
                    command = "curl -X {method} -H {headers} -d '{uri}'"
                    headers = ['"{0}: {1}"'.format(k, v) for k, v in resp.request_info.headers.items()]
                    headers = " -H ".join(headers)
                    self.last_curl = command.format(method=resp.request_info.method,
                                                    headers=headers,
                                                    uri=resp.request_info.url)
                    text = await resp.text()
                    try:
                        _json = await resp.json()
                    except:
                        _json = dict()
                    response = AsyncResponse(text=text,
                                             _json=_json,
                                             async_resp=resp)
            except Exception as err:
                response = AsyncResponseError(error=err, trace=traceback.format_exc())
            finally:
                if pbar is not None:
                    pbar.close()
                with threadLock:
                    self.calls_counter.add()
        if response.ok and self.cache is not None:
            try:
                self.cache.write(list_entities_json=[response.json()])
                dataset_id = url.split('/')[-2]
                self.cache.write_stream(request_path=url,
                                        buffer=to_upload,
                                        file_name=uploaded_filename,
                                        entity_id=response.json()['id'],
                                        dataset_id=dataset_id)
            except:
                logger.warning("Failed to add the file to the cache")
        return response

    def __get_pbar(self, pbar, total_length):
        # decide if create progress bar for item
        if pbar:
            try:
                if total_length is not None and int(total_length) > 10e6:  # size larger than 10 MB:
                    pbar = tqdm.tqdm(total=int(total_length),
                                     unit='B',
                                     unit_scale=True,
                                     unit_divisor=1024,
                                     position=1,
                                     file=sys.stdout,
                                     disable=self.verbose.disable_progress_bar)
                else:
                    pbar = None
            except Exception as err:
                pbar = None
                logger.debug('Cant decide downloaded file length, bar will not be presented: {}'.format(err))
        return pbar

    def send_session(self, prepared, stream=None):
        if self.session is None:
            self.session = requests.Session()
            retry = Retry(
                total=5,
                read=5,
                connect=5,
                backoff_factor=1,
                # use on any request type
                allowed_methods=False,
                # force retry on those status responses
                status_forcelist=(501, 502, 503, 504, 505, 506, 507, 508, 510, 511),
                raise_on_status=False
            )
            adapter = HTTPAdapter(max_retries=retry,
                                  pool_maxsize=np.sum(list(self._thread_pools_names.values())),
                                  pool_connections=np.sum(list(self._thread_pools_names.values())))
            self.session.mount('http://', adapter)
            self.session.mount('https://', adapter)
        resp = self.session.send(request=prepared, stream=stream, verify=self.verify, timeout=120)

        with threadLock:
            self.calls_counter.add()

        return resp

    @staticmethod
    def check_proxy():
        """
        Verify that dataloop urls are not blocked
        :return:
        """
        proxy_envs = ['HTTP', 'HTTPS', 'http', 'https']
        dataloop_urls = ['dev-gate.dataloop.ai',
                         'gate.dataloop.ai',
                         'dataloop-development.auth0.com',
                         'dataloop-production.auth0.com']
        if True in [env in os.environ for env in proxy_envs]:
            # check if proxy exists
            if True in [env in os.environ for env in ['no_proxy', 'NO_PROXY']]:
                # check if no_proxy exists
                if 'no_proxy' in os.environ:
                    # check if dataloop urls in no_proxy
                    if True not in [url in os.environ['no_proxy'] for url in dataloop_urls]:
                        # no dataloop url exists in no_proxy
                        logger.warning('Proxy is used, make sure dataloop urls are in "no_proxy" environment variable')
                else:
                    # check if dataloop urls in no_proxy
                    if True not in [url in os.environ['NO_PROXY'] for url in dataloop_urls]:
                        # no dataloop url exists in no_proxy
                        logger.warning('Proxy is used, make sure dataloop urls are in "no_proxy" environment variable')
            else:
                logger.warning('Proxy is used, make sure dataloop urls are in "no_proxy" environment variable')

    def token_expired(self, t=60):
        """
        Check token validation
        :param t: time ahead interval in seconds
        """
        try:
            if self.token is None or self.token == '':
                expired = True
            else:
                # oxsec-disable jwt-signature-disabled - Client-side SDK: signature verification disabled intentionally to check token expiration; server validates on API calls
                payload = jwt.decode(
                    self.token,
                    options={
                        "verify_signature": False,
                        "verify_exp": False,
                        "verify_aud": False,
                        "verify_iss": False,
                    }
                )
                d = datetime.datetime.now(datetime.timezone.utc)
                epoch = datetime.datetime(1970, 1, 1, tzinfo=datetime.timezone.utc)
                now = (d - epoch).total_seconds()
                exp = payload['exp']
                if now < (exp - t):
                    expired = False
                else:
                    expired = True
        except jwt.exceptions.DecodeError:
            logger.exception('Invalid token.')
            expired = True
        except Exception:
            logger.exception('Unknown error:')
            expired = True
        if expired:
            if self.renew_token_method():
                expired = False
        return expired

    @staticmethod
    def is_json_serializable(response):
        try:
            response_json = response.json()
            return True, response_json
        except ValueError:
            return False, None

    ##########
    # STDOUT #
    ##########
    def print_response(self, resp=None):
        """
        Print tabulate response
        :param resp: response from requests
        :return:
        """
        try:
            if resp is None:
                resp = self.last_response
            is_json_serializable, results = self.is_json_serializable(response=resp)
            if self.verbose.print_all_responses and is_json_serializable:
                if isinstance(results, dict):
                    to_print = miscellaneous.List([results])
                elif isinstance(results, list):
                    to_print = miscellaneous.List(results)
                else:
                    logger.debug('Unknown response type: {}. cant print'.format(type(results)))
                    return
                request_id = resp.headers.get('x-request-id', 'na')
                logger.debug('--- [Request] Start ---')
                logger.debug(self.print_request(req=resp.request, to_return=True))
                logger.debug('--- [Request] End ---')
                logger.debug('--- [Response][x-request-id:{}] Start ---'.format(request_id))
                to_print.print(show_all=False, level='debug')
                logger.debug('--- [Response][x-request-id:{}] End ---'.format(request_id))
        except Exception:
            logger.exception('Printing response from gate:')

    def print_bad_response(self, resp=None, log_error=True):
        """
        Print error from platform
        :param resp:
        :param log_error: print error log (to use when trying request more than once)
        :return:
        """
        if resp is None:
            resp = self.last_response
        msg = ''
        if hasattr(resp, 'status_code'):
            msg += '[Response <{val}>]'.format(val=resp.status_code)
        if hasattr(resp, 'reason'):
            msg += '[Reason: {val}]'.format(val=resp.reason)
        if hasattr(resp, 'text') and isinstance(resp.text, str):
            msg += '[Text: {val}]'.format(val=format_message(resp.text))

        request_id = resp.headers.get('x-request-id', 'na')
        logger.debug('--- [Request] Start ---')
        logger.debug(self.print_request(req=resp.request, to_return=True))
        logger.debug('--- [Request] End ---')
        logger.debug('--- [Response][x-request-id:{}] Start ---'.format(request_id))
        if log_error:
            logger.error(msg)
        else:
            logger.debug(msg)
        logger.debug('--- [Response][x-request-id:{}] End ---'.format(request_id))
        self.platform_exception = PlatformError(resp)

    def print_request(self, req=None, to_return=False, with_auth=False):
        """
        Print a request to the platform
        :param req:
        :param to_return: return string instead of printing
        :param with_auth: print authentication
        :return:
        """
        if not req:
            req = self.last_request

        headers = list()
        for k, v in req.headers.items():
            if k == 'authorization' and not with_auth:
                continue
            headers.append('{}: {}'.format(k, v))
        if hasattr(req, 'body'):
            body = req.body
        elif isinstance(req, aiohttp.RequestInfo):
            body = {'multipart': 'true'}
        else:
            body = dict()

        # remove secrets and passwords
        try:
            body = json.loads(body)
            if isinstance(body, dict):
                for key, value in body.items():
                    hide = any([field in key for field in ['secret', 'password']])
                    if hide:
                        body[key] = '*' * len(value)
        except Exception:
            pass

        msg = '{}\n{}\n{}'.format(
            req.method + ' ' + str(req.url),
            '\n'.join(headers),
            body,
        )
        if to_return:
            return msg
        else:
            print(msg)

    ################
    # Environments #
    ################
    def setenv(self, env):
        """
        Set environment
        :param env:
        :return:
        """

        environments = self.environments
        if env.startswith('http'):
            if env not in environments.keys():
                msg = 'Unknown environment. Please add environment to SDK ("add_environment" method)'
                logger.error(msg)
                raise ConnectionError(msg)
        elif env == 'custom':
            custom_env = os.environ.get('DTLPY_CUSTOM_ENV', None)
            environment = json.loads(base64.b64decode(custom_env.encode()).decode())
            env = environment.pop('url')
            token = None
            if self.environments.get(env):
                token = self.environments[env].get('token', None)
            self.environments[env] = environment.get(env, environment)
            self.environments[env]['token'] = token
            verify_ssl = self.environments[env].get('verify_ssl', None)
            if verify_ssl is not None and isinstance(verify_ssl, str):
                self.environments[env]['verify_ssl'] = True if verify_ssl.lower() == 'true' else False
        else:
            matched_env = [env_url for env_url, env_dict in environments.items() if env_dict['alias'] == env]
            if len(matched_env) != 1:
                known_aliases = [env_dict['alias'] for env_url, env_dict in environments.items()]
                raise ConnectionError(
                    'Unknown platform environment: "{}". Known: {}'.format(env, ', '.join(known_aliases)))
            env = matched_env[0]
        if self.environment != env:
            self.environment = env
            self.__gate_url_for_requests = None
            # reset local token
            self._token = None
            self.refresh_token_active = True
        logger.info('Platform environment: {}'.format(self.environment))
        if self.token_expired():
            logger.info('Token expired, Please login.')

    ##########
    # Log in #
    ##########
    def login_secret(self, email, password, client_id, client_secret=None, force=False):
        """
        Login with email and password from environment variables.
        If already logged in with same user - login will NOT happen. see "force"

        :param email: user email.
        :param password: user password
        :param client_id: auth0 client id
        :param client_secret: secret that match the client id
        :param force: force login. in case login with same user but want to get a new JWT
        :return:
        """
        logger.warning('dl.login_secret is deprecated. Please use dl.login_m2m instead.')
        return login_secret(api_client=self,
                            email=email,
                            password=password,
                            client_id=client_id,
                            client_secret=client_secret,
                            force=force)

    def login_m2m(self, email, password, client_id=None, client_secret=None, force=False):
        """
        Login with email and password from environment variables
        :param email: user email. if already logged in with same user - login will NOT happen. see "force"
        :param password: user password
        :param client_id:
        :param client_secret:
        :param force: force login. in case login with same user but want to get a new JWT
        :return:
        """
        res = login_m2m(api_client=self,
                        email=email,
                        password=password,
                        client_id=client_id,
                        client_secret=client_secret,
                        force=force)
        # if res:
        #     self._send_login_event(user_type='human', login_type='m2m')
        return res

    def login_token(self, token):
        """
        Login using existing token
        :param token: a valid token
        :return:
        """
        current_token = self.token
        self.token = token
        success, response = self.gen_request(req_type='get', path='/users/me')
        if not response.ok:
            # switch back to before
            self.token = current_token
            raise ValueError(f"Invalid API key provided. Error: {response.text}")

    def login_api_key(self, api_key):
        """
        Login using API key
        :param api_key: a valid API key
        :return:
        """
        current_token = self.token
        self.token = api_key
        success, response = self.gen_request(req_type='get', path='/users/me')
        if not response.ok:
            # switch back to before
            self.token = current_token
            raise ValueError(f"Invalid API key provided. Error: {response.text}")

    @property
    def login_domain(self):
        if self._login_domain is None:
            self._login_domain = self.environments[self.environment].get('login_domain', None)
        return self._login_domain

    @login_domain.setter
    def login_domain(self, domain: str):
        if domain is not None and not isinstance(domain, str):
            raise exceptions.PlatformException('400', 'domain should be a string value')
        self._login_domain = domain
        self.environments[self.environment]['login_domain'] = domain
        self.cookie_io.put('login_parameters', self.environments)

    def login(self, audience=None, auth0_url=None, client_id=None, callback_port=None):
        """
        Login using Auth0.
        :return:
        """
        res = login(
            api_client=self,
            audience=audience,
            auth0_url=auth0_url,
            client_id=client_id,
            login_domain=self.login_domain,
            callback_port=callback_port
        )
        # if res:
        #     self._send_login_event(user_type='human', login_type='interactive')
        return res

    # def _send_login_event(self, user_type, login_type):
    #     event_payload = {
    #         'event': 'dtlpy:login',
    #         'properties': {
    #             'login_type': login_type,
    #             'user_type': user_type
    #         }
    #     }
    #     self.event_tracker.put(event=event_payload)

    def logout(self):
        """
        Logout.
        :return:
        """
        return logout(api_client=self)

    def _renew_token_in_dual_agent(self):
        renewed = False
        try:
            proxy_port = os.environ.get('AGENT_PROXY_MAIN_PORT') or "1001"
            resp = requests.get('http://localhost:{port}/get_jwt'.format(port=proxy_port))
            if resp.ok:
                self.token = resp.json()['jwt']
                renewed = True
            else:
                self.print_bad_response(resp)
        except Exception:
            logger.exception('Failed to get token from proxy')

        return renewed

    def renew_token(self):
        refresh_method = os.environ.get('DTLPY_REFRESH_TOKEN_METHOD', None)
        if refresh_method is not None and refresh_method == 'proxy':
            res = self._renew_token_in_dual_agent()
        else:
            res = self._renew_token_with_refresh_token()
        # if res:
        #     self._send_login_event(user_type='human', login_type='refresh')
        return res

    def generate_api_key(self, description: str = None, login: bool = False):
        """
        Generate an API key for a user
        :param description: description for the API key
        :param login: if True, login with the new API key
        :return: User token
        """
        user_email = self.info()['user_email']
        payload = {
            'userId': user_email
        }
        if description:
            if not isinstance(description, str):
                raise ValueError('description should be a string')
            payload['description'] = description
        success, response = self.gen_request(req_type='post', path='/apiKeys', json_req=payload)
        if not success:
            raise exceptions.PlatformException(response)
        if login:
            self.login_api_key(response.json()['jwt'])
            return True

        return response.json()['jwt']

    def _renew_token_with_refresh_token(self):
        renewed = False
        if self.refresh_token_active is False:
            return renewed
        logger.debug('RefreshToken: Started')
        if self.token is None or self.token == '':
            # token is missing
            logger.debug('RefreshToken: Missing token.')
            self.refresh_token_active = False
        if self.refresh_token is None or self.refresh_token == '':
            # missing refresh token
            logger.debug('RefreshToken: Missing "refresh_token"')
            self.refresh_token_active = False
        if self.environment not in self.environments.keys():
            # env params missing
            logger.debug('RefreshToken: Missing environments params for refreshing token')
            self.refresh_token_active = False

        if self.refresh_token_active is False:
            return renewed

        refresh_token = self.refresh_token

        env_params = self.environments[self.environment]
        if 'gate_url' not in env_params:
            env_params['gate_url'] = gate_url_from_host(environment=self.environment)
            self.environments[self.environment] = env_params
        token_endpoint = "{}/token?default".format(env_params['gate_url'])

        payload = {
            'type': 'refresh_token',
            'refresh_token': refresh_token
        }
        logger.debug("RefreshToken: Refreshing token via {}".format(token_endpoint))
        resp = requests.request(
            "POST",
            token_endpoint,
            json=payload,
            headers={'content-type': 'application/json'},
            verify=self.verify
        )
        if not resp.ok:
            logger.debug('RefreshToken: Failed')
            self.print_bad_response(resp)
        else:
            response_dict = resp.json()
            # get new token
            final_token = response_dict['id_token']
            self.token = final_token
            self.refresh_token = refresh_token
            # set status back to pending
            logger.debug('RefreshToken: Success')
            renewed = True
        return renewed

    def set_api_counter(self, filepath):
        self.calls_counter = CallsCounter(filepath=filepath)

    def _get_resource_url(self, url):

        env = self._environments[self._environment]['alias']
        head = self._environments[self._environment].get('url', None)
        # TODO need to deprecate somehow (the following)
        if head is None:
            if env == 'prod':
                head = 'https://console.dataloop.ai/'
            elif env == 'dev':
                head = 'https://dev-con.dataloop.ai/'
            elif env == 'rc':
                head = 'https://rc-con.dataloop.ai/'
            elif env in ['local', 'minikube_local_mac']:
                head = 'https://localhost:8443/'
            elif env == 'new-dev':
                head = 'https://custom1-gate.dataloop.ai/'
            else:
                raise exceptions.PlatformException(error='400', message='Unknown environment: {}'.format(env))

        return head + url

    def _open_in_web(self, url):
        import webbrowser
        webbrowser.open(url=url, new=2, autoraise=True)


client = ApiClient()



def run():
    """
    Play a video from platform with annotations in the Dataloop python player
    :return:
    """
    from dtlpy.utilities.videos.video_player import VideoPlayer
    project_name = 'Dancing'
    dataset_name = 'Flossing'
    item_name = '/first_try.mp4'

    VideoPlayer(project_name=project_name,
                dataset_name=dataset_name,
                item_filepath=item_name)



class JsonUtils:
    @staticmethod
    def get_if_absent(value, default=None):
        if value is None:
            if default is None:
                value_type = type(value)
                if value_type == dict:
                    default = {}
                elif value_type == list:
                    default = []
                else:
                    default = ''
            value = default
        return value



import attr
import traceback
from enum import Enum
from collections import namedtuple

from .. import repositories, entities
from ..services.api_client import ApiClient


class CommandsStatus(str, Enum):
    CREATED = 'created'
    MAKING_CHILDREN = 'making-children'
    WAITING_CHILDREN = 'waiting-children'
    IN_PROGRESS = 'in-progress'
    ABORTED = 'aborted'
    CANCELED = 'canceled'
    FINALIZING = 'finalizing'
    SUCCESS = 'success'
    FAILED = 'failed'
    TIMEOUT = 'timeout'
    CLEANING_UP = 'cleaning-up'
    ON_ERROR = 'on-error'


@attr.s
class Command(entities.BaseEntity):
    """
    Com entity
    """
    # platform
    id = attr.ib()
    url = attr.ib(repr=False)
    status = attr.ib()
    created_at = attr.ib()
    updated_at = attr.ib()
    type = attr.ib()
    progress = attr.ib()
    spec = attr.ib()
    error = attr.ib()

    # sdk
    _client_api = attr.ib(type=ApiClient, repr=False)
    _repositories = attr.ib(repr=False)

    ################
    # repositories #
    ################
    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['commands'])

        commands_repo = repositories.Commands(client_api=self._client_api)

        r = reps(commands=commands_repo)
        return r

    @property
    def commands(self):
        assert isinstance(self._repositories.commands, repositories.Commands)
        return self._repositories.commands

    @staticmethod
    def _protected_from_json(_json, client_api, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error

        :param _json: platform json
        :param client_api: ApiClient entity
        :param is_fetched: is Entity fetched from Platform
        :return:
        """
        try:
            command = Command.from_json(_json=_json,
                                        client_api=client_api,
                                        is_fetched=is_fetched)
            status = True
        except Exception:
            command = traceback.format_exc()
            status = False
        return status, command

    @classmethod
    def from_json(cls, _json, client_api, is_fetched=True):
        """
        Build a Command entity object from a json

        :param _json: _json response from host
        :param client_api: ApiClient entity
        :param is_fetched: is Entity fetched from Platform
        :return: Command object
        """
        inst = cls(
            id=_json.get('id', None),
            url=_json.get('url', None),
            status=_json.get('status', None),
            created_at=_json.get('createdAt', None),
            updated_at=_json.get('updatedAt', None),
            type=_json.get('type', None),
            progress=_json.get('progress', None),
            spec=_json.get('spec', None),
            error=_json.get('error', None),
            client_api=client_api
        )
        inst.is_fetched = is_fetched
        return inst

    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        # get excluded
        _json = attr.asdict(self, filter=attr.filters.exclude(attr.fields(Command)._client_api,
                                                              attr.fields(Command).created_at,
                                                              attr.fields(Command).updated_at))
        # rename
        _json['createdAt'] = self.created_at
        _json['updatedAt'] = self.updated_at

        return _json

    def abort(self):
        """
        abort command

        :return:
        """
        return self.commands.abort(command_id=self.id)

    def in_progress(self):
        """
        Check if command is still in one of the in progress statuses

        :return: True if command still in progress
        :rtype: bool
        """
        if self.status not in {status for status in entities.CommandsStatus}:
            raise ValueError('status is not a valid CommandsStatus')
        return self.status not in [entities.CommandsStatus.SUCCESS,
                                   entities.CommandsStatus.FAILED,
                                   entities.CommandsStatus.TIMEOUT,
                                   entities.CommandsStatus.CANCELED,
                                   entities.CommandsStatus.ABORTED
                                   ]

    def wait(self, timeout=0, step=None, backoff_factor=1, iteration_callback=None):
        """
        Wait for Command to finish

        :param int timeout: int, seconds to wait until TimeoutError is raised. if 0 - wait until done
        :param int step: int, seconds between polling
        :param float backoff_factor: A backoff factor to apply between attempts after the second try
        :param function iteration_callback: function to call on each iteration
        :return: Command  object
        """
        if not self.in_progress():
            return self

        return self.commands.wait(
            command_id=self.id,
            timeout=timeout,
            step=step,
            url=self.url,
            backoff_factor=backoff_factor,
            iteration_callback=iteration_callback
        )



import dtlpy as dl


class ModelAdapter(dl.BaseModelAdapter):
    """
    Specific Model adapter.
    The class bind a dl.Model entity with the package code
    """
    # TODO:
    #   1) docstring for your ModelAdapter
    #   2) implement the virtual methods for full adapter support
    #   3) add your _defaults

    _defaults = {}

    def __init__(self, model_entity):
        super(ModelAdapter, self).__init__(model_entity)

    # ===============================
    # NEED TO IMPLEMENT THESE METHODS
    # ===============================

    def load(self, local_path, **kwargs):
        """ Loads model and populates self.model with a `runnable` model

            Virtual method - need to implement

            This function is called by load_from_model (download to local and then loads)

        :param local_path: `str` directory path in local FileSystem
        """
        raise NotImplementedError("Please implement 'load' method in {}".format(self.__class__.__name__))

    def save(self, local_path, **kwargs):
        """ saves configuration and weights locally

            Virtual method - need to implement

            the function is called in save_to_model which first save locally and then uploads to model entity

        :param local_path: `str` directory path in local FileSystem
        """
        raise NotImplementedError("Please implement 'save' method in {}".format(self.__class__.__name__))

    def train(self, data_path, dump_path, **kwargs):
        """
        Virtual method - need to implement
        Train the model according to data in local_path and save everything to dump_path

        :param data_path: `str` local File System path to where the data was downloaded and converted at
        :param dump_path: `str` local File System path where to dump training mid-results (checkpoints, logs...)
        """
        raise NotImplementedError("Please implement 'train' method in {}".format(self.__class__.__name__))

    def predict(self, batch, **kwargs):
        """ Model inference (predictions) on batch of images

            Virtual method - need to implement

        :param batch: `np.ndarray`
        :return: `list[dl.AnnotationCollection]` each collection is per each image / item in the batch
        """
        raise NotImplementedError("Please implement 'predict' method in {}".format(self.__class__.__name__))

    def convert(self, data_path, **kwargs):
        """ Convert Dataloop structure data to model structured

            Virtual method - need to implement

            e.g. take dlp dir structure and construct annotation file

        :param data_path: `str` local File System directory path where
                           we already downloaded the data from dataloop platform
        :return:
        """
        raise NotImplementedError("Please implement 'convert' method in {}".format(self.__class__.__name__))

    # NOT IN USE
    def convert_dlp(self, items: dl.entities.PagedEntities):
        """ This should implement similar to convert only to work on dlp items.  ->
                   -> meaning create the converted version from items entities"""
        # TODO
        pass



import logging

from .dataset_generator import DatasetGenerator

logger = logging.getLogger(name='dtlpy')

try:
    import tensorflow
except (ImportError, ModuleNotFoundError):
    logger.error('Failed importing tensorflow package, cannot use DatasetGeneratorTensorflow')


class DatasetGeneratorTensorflow(DatasetGenerator, tensorflow.keras.utils.Sequence):
    def __getitem__(self, item):
        batch = super(DatasetGeneratorTensorflow, self).__getitem__(item)
        x = batch['image']
        y = batch['annotations']
        return x, y

    def __iter__(self):
        """Create a generator that iterate over the Sequence."""
        for item in (self[i] for i in range(len(self))):
            yield item



from . import BaseAnnotationDefinition


class FreeText(BaseAnnotationDefinition):
    """
    Free text annotation type (e.g. response for a prompt)
    """
    type = "text"

    def __init__(self, text, label='free-text', attributes=None, description=None):
        """
        Create a free text annotation
        :param label: annotation label
        :param text: string of the annotation
        :param attributes: annotation attributes
        :param description:

        :return:
        """
        super().__init__(description=description, attributes=attributes)
        self.text = text
        self.label = label

    @property
    def x(self):
        return 0

    @property
    def y(self):
        return 0

    @property
    def geo(self):
        return list()

    def show(self, image, thickness, with_text, height, width, annotation_format, color, alpha=1):
        """
        Show annotation as ndarray
        :param image: empty or image to draw on
        :param thickness:
        :param with_text: not required
        :param height: item height
        :param width: item width
        :param annotation_format: options: list(dl.ViewAnnotationOptions)
        :param color: color
        :param alpha: opacity value [0 1], default 1
        :return: ndarray
        """
        return self.text

    def to_coordinates(self, color):
        return self.text

    @classmethod
    def from_json(cls, _json):
        coordinates = _json["coordinates"]

        return cls(
            text=coordinates,
            label=_json["label"],
            attributes=_json.get("attributes", None),
        )



import warnings
from collections import namedtuple
from enum import Enum
import traceback
import logging
from typing import List
from urllib.parse import urlsplit
import attr
from .. import repositories, entities
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')

class DynamicConcurrencyUpdateMethod(str, Enum):
    """ The method of updating the dynamic concurrency.
    """
    RESTART = 'restart',
    SYNC = 'sync'


class ServiceType(str, Enum):
    """ The type of the service (SYSTEM).

    .. list-table::
       :widths: 15 150
       :header-rows: 1

       * - State
         - Description
       * - SYSTEM
         - Dataloop internal service
    """
    SYSTEM = 'system'
    REGULAR = 'regular'


class ServiceModeType(str, Enum):
    """ The type of the service mode.

    .. list-table::
       :widths: 15 150
       :header-rows: 1

       * - State
         - Description
       * - REGULAR
         - Service regular mode type
       * - DEBUG
         - Service debug mode type
    """
    REGULAR = 'regular'
    DEBUG = 'debug'


class OnResetAction(str, Enum):
    """ The Execution action when the service reset (RERUN, FAILED).

    .. list-table::
       :widths: 15 150
       :header-rows: 1

       * - State
         - Description
       * - RERUN
         - When the service resting rerun the execution
       * - FAILED
         - When the service resting fail the execution
    """
    RERUN = 'rerun'
    FAILED = 'failed'


class InstanceCatalog(str, Enum):
    """ The Service Pode size.

    .. list-table::
       :widths: 15 150
       :header-rows: 1

       * - State
         - Description
       * - REGULAR_XS
         - regular pod with extra small size
       * - REGULAR_S
         - regular pod with small size
       * - REGULAR_M
         - regular pod with medium size
       * - REGULAR_L
         - regular pod with large size
       * - HIGHMEM_XS
         - highmem pod with extra small size
       * - HIGHMEM_S
         - highmem pod with small size
       * - HIGHMEM_M
         - highmem pod with medium size
       * - HIGHMEM_L
         - highmem pod with large size
       * - GPU_T4_S
         - GPU NVIDIA T4 pod with regular memory
       * - GPU_T4_M
         - GPU NVIDIA T4 pod with highmem
    """
    REGULAR_XS = "regular-xs"
    REGULAR_S = "regular-s"
    REGULAR_M = "regular-m"
    REGULAR_L = "regular-l"
    HIGHMEM_XS = "highmem-xs"
    HIGHMEM_S = "highmem-s"
    HIGHMEM_M = "highmem-m"
    HIGHMEM_L = "highmem-l"
    GPU_T4_S = "gpu-t4"
    GPU_T4_M = "gpu-t4-m"


class RuntimeType(str, Enum):
    """ Service culture Runtime (KUBERNETES).

    .. list-table::
       :widths: 15 150
       :header-rows: 1

       * - State
         - Description
       * - KUBERNETES
         - Service run in kubernetes culture
    """
    KUBERNETES = 'kubernetes'


class ServiceRuntime(entities.BaseEntity):
    def __init__(self, service_type: RuntimeType = RuntimeType.KUBERNETES):
        self.service_type = service_type


class KubernetesRuntime(ServiceRuntime):
    DEFAULT_POD_TYPE = InstanceCatalog.REGULAR_S
    DEFAULT_NUM_REPLICAS = 1
    DEFAULT_CONCURRENCY = 10

    def __init__(self,
                 pod_type: InstanceCatalog = DEFAULT_POD_TYPE,
                 num_replicas=DEFAULT_NUM_REPLICAS,
                 concurrency=DEFAULT_CONCURRENCY,
                 dynamic_concurrency=None,
                 dynamic_concurrency_config=None,
                 runner_image=None,
                 autoscaler=None,
                 **kwargs):

        super().__init__(service_type=RuntimeType.KUBERNETES)
        self.pod_type = kwargs.get('podType', pod_type)
        self.num_replicas = kwargs.get('numReplicas', num_replicas)
        self.concurrency = kwargs.get('concurrency', concurrency)
        self.runner_image = kwargs.get('runnerImage', runner_image)
        self._proxy_image = kwargs.get('proxyImage', None)
        self.single_agent = kwargs.get('singleAgent', None)
        self.preemptible = kwargs.get('preemptible', None)
        self.dynamic_concurrency = kwargs.get('dynamicConcurrency', dynamic_concurrency)
        self.dynamic_concurrency_config = kwargs.get('dynamicConcurrencyConfig', dynamic_concurrency_config)

        self.autoscaler = kwargs.get('autoscaler', autoscaler)
        if self.autoscaler is not None and isinstance(self.autoscaler, dict):
            if self.autoscaler['type'] == KubernetesAutoscalerType.RABBITMQ:
                self.autoscaler = KubernetesRabbitmqAutoscaler(**self.autoscaler)
            elif self.autoscaler['type'] == KubernetesAutoscalerType.RPS:
                self.autoscaler = KubernetesRPSAutoscaler(**self.autoscaler)
            else:
                raise NotImplementedError(
                    'Unknown kubernetes autoscaler type: {}'.format(self.autoscaler['type']))

    def to_json(self):
        _json = {
            'podType': self.pod_type,
            'numReplicas': self.num_replicas,
            'concurrency': self.concurrency,
            'autoscaler': None if self.autoscaler is None else self.autoscaler.to_json()
        }

        if self.single_agent is not None:
            _json['singleAgent'] = self.single_agent

        if self.runner_image is not None:
            _json['runnerImage'] = self.runner_image

        if self._proxy_image is not None:
            _json['proxyImage'] = self._proxy_image

        if self.preemptible is not None:
            _json['preemptible'] = self.preemptible

        if self.dynamic_concurrency is not None:
            _json['dynamicConcurrency'] = self.dynamic_concurrency

        if self.dynamic_concurrency_config is not None:
            _json['dynamicConcurrencyConfig'] = self.dynamic_concurrency_config

        return _json


@attr.s
class Service(entities.BaseEntity):
    """
    Service object
    """
    # platform
    created_at = attr.ib()
    updated_at = attr.ib(repr=False)
    creator = attr.ib()
    version = attr.ib()

    package_id = attr.ib()
    package_revision = attr.ib()

    bot = attr.ib()
    use_user_jwt = attr.ib(repr=False)
    init_input = attr.ib()
    versions = attr.ib(repr=False)
    module_name = attr.ib()
    name = attr.ib()
    url = attr.ib()
    id = attr.ib()
    active = attr.ib()
    driver_id = attr.ib(repr=False)
    secrets = attr.ib(repr=False)

    # name change
    runtime = attr.ib(repr=False, type=KubernetesRuntime)
    queue_length_limit = attr.ib()
    run_execution_as_process = attr.ib(type=bool)
    execution_timeout = attr.ib()
    drain_time = attr.ib()
    on_reset = attr.ib(type=OnResetAction)
    _type = attr.ib(type=ServiceType)
    project_id = attr.ib()
    org_id = attr.ib()
    is_global = attr.ib()
    max_attempts = attr.ib()
    mode = attr.ib(repr=False)
    metadata = attr.ib()
    archive = attr.ib(repr=False)
    config = attr.ib(repr=False)
    settings = attr.ib(repr=False)
    panels = attr.ib(repr=False)

    # SDK
    _package = attr.ib(repr=False)
    _client_api = attr.ib(type=ApiClient, repr=False)
    _revisions = attr.ib(default=None, repr=False)
    # repositories
    _project = attr.ib(default=None, repr=False)
    _repositories = attr.ib(repr=False)
    updated_by = attr.ib(default=None)
    app = attr.ib(default=None)
    integrations = attr.ib(default=None)

    @property
    def createdAt(self):
        return self.created_at

    @property
    def updatedAt(self):
        return self.updated_at

    @staticmethod
    def _protected_from_json(_json: dict, client_api: ApiClient, package=None, project=None, is_fetched=True):
        """
        Same as from_json but with try-except to catch if error

        :param _json: platform json
        :param client_api: ApiClient entity
        :param package:
        :param project: project entity
        :param is_fetched: is Entity fetched from Platform
        :return:
        """
        try:
            service = Service.from_json(_json=_json,
                                        client_api=client_api,
                                        package=package,
                                        project=project,
                                        is_fetched=is_fetched)
            status = True
        except Exception:
            service = traceback.format_exc()
            status = False
        return status, service

    @classmethod
    def from_json(cls, _json: dict, client_api: ApiClient = None, package=None, project=None, is_fetched=True):
        """
        Build a service entity object from a json

        :param dict _json: platform json
        :param dl.ApiClient client_api: ApiClient entity
        :param dtlpy.entities.package.Package package: package entity
        :param dtlpy.entities.project.Project project: project entity
        :param bool is_fetched: is Entity fetched from Platform
        :return: service object
        :rtype: dtlpy.entities.service.Service
        """
        if project is not None:
            if project.id != _json.get('projectId', None):
                logger.warning('Service has been fetched from a project that is not belong to it')
                project = None

        if package is not None:
            if package.id != _json.get('packageId', None):
                logger.warning('Service has been fetched from a package that is not belong to it')
                package = None

        versions = _json.get('versions', dict())
        runtime = _json.get("runtime", None)
        if runtime:
            runtime = KubernetesRuntime(**runtime)

        inst = cls(
            package_revision=_json.get("packageRevision", None),
            bot=_json.get("botUserName", None),
            use_user_jwt=_json.get("useUserJwt", False),
            created_at=_json.get("createdAt", None),
            updated_at=_json.get("updatedAt", None),
            project_id=_json.get('projectId', None),
            package_id=_json.get('packageId', None),
            driver_id=_json.get('driverId', None),
            max_attempts=_json.get('maxAttempts', None),
            version=_json.get('version', None),
            creator=_json.get('creator', None),
            revisions=_json.get('revisions', None),
            queue_length_limit=_json.get('queueLengthLimit', None),
            active=_json.get('active', None),
            runtime=runtime,
            is_global=_json.get("global", False),
            init_input=_json.get("initParams", dict()),
            module_name=_json.get("moduleName", None),
            run_execution_as_process=_json.get('runExecutionAsProcess', False),
            execution_timeout=_json.get('executionTimeout', 60 * 60),
            drain_time=_json.get('drainTime', 60 * 10),
            on_reset=_json.get('onReset', OnResetAction.FAILED),
            name=_json.get("name", None),
            url=_json.get("url", None),
            id=_json.get("id", None),
            versions=versions,
            client_api=client_api,
            package=package,
            project=project,
            secrets=_json.get("secrets", None),
            type=_json.get("type", None),
            mode=_json.get('mode', dict()),
            metadata=_json.get('metadata', None),
            archive=_json.get('archive', None),
            updated_by=_json.get('updatedBy', None),
            config=_json.get('config', None),
            settings=_json.get('settings', None),
            app=_json.get('app', None),
            integrations=_json.get('integrations', None),
            org_id=_json.get('orgId', None),
            panels=_json.get('panels', None)
        )
        inst.is_fetched = is_fetched
        return inst

    ############
    # Entities #
    ############
    @property
    def revisions(self):
        if self._revisions is None:
            self._revisions = self.services.revisions(service=self)
        return self._revisions

    @property
    def platform_url(self):
        return self._client_api._get_resource_url("projects/{}/services/{}/main".format(self.project.id, self.id))

    @property
    def project(self):
        if self._project is None:
            self._project = repositories.Projects(client_api=self._client_api).get(project_id=self.project_id,
                                                                                   fetch=None)
        assert isinstance(self._project, entities.Project)
        return self._project

    @property
    def package(self):
        if self._package is None:
            try:
                dpk_id = None
                dpk_version = None
                if self.app and isinstance(self.app, dict):
                    dpk_id = self.app.get('dpkId', None)
                    dpk_version = self.app.get('dpkVersion', None)
                if dpk_id is None:
                    self._package = repositories.Dpks(client_api=self._client_api, project=self.project).get(
                        dpk_id=self.package_id)
                else:
                    self._package = repositories.Dpks(client_api=self._client_api, project=self.project).get_revisions(
                        dpk_id=dpk_id,
                        version=dpk_version)

                assert isinstance(self._package, entities.Dpk)
            except:
                self._package = repositories.Packages(client_api=self._client_api).get(package_id=self.package_id,
                                                                                       fetch=None,
                                                                                       log_error=False)
                assert isinstance(self._package, entities.Package)
        return self._package

    @property
    def execution_url(self):
        return 'CURL -X POST' \
               '\nauthorization: Bearer <token>' \
               '\nContent-Type: application/json" -d {' \
               '\n"input": {<input json>}, ' \
               '"projectId": "{<project_id>}", ' \
               '"functionName": "<function_name>"}'

    ################
    # repositories #
    ################
    @_repositories.default
    def set_repositories(self):
        reps = namedtuple('repositories',
                          field_names=['executions', 'services', 'triggers'])

        if self._package is None:
            services_repo = repositories.Services(client_api=self._client_api,
                                                  package=self._package,
                                                  project=self._project)
        else:
            services_repo = self._package.services

        triggers = repositories.Triggers(client_api=self._client_api,
                                         project=self._project,
                                         service=self)

        r = reps(executions=repositories.Executions(client_api=self._client_api, service=self),
                 services=services_repo, triggers=triggers)
        return r

    @property
    def executions(self):
        assert isinstance(self._repositories.executions, repositories.Executions)
        return self._repositories.executions

    @property
    def triggers(self):
        assert isinstance(self._repositories.triggers, repositories.Triggers)
        return self._repositories.triggers

    @property
    def services(self):
        assert isinstance(self._repositories.services, repositories.Services)
        return self._repositories.services

    ###########
    # methods #
    ###########
    def to_json(self):
        """
        Returns platform _json format of object

        :return: platform json format of object
        :rtype: dict
        """
        _json = attr.asdict(
            self,
            filter=attr.filters.exclude(
                attr.fields(Service)._project,
                attr.fields(Service)._package,
                attr.fields(Service)._revisions,
                attr.fields(Service)._client_api,
                attr.fields(Service)._repositories,
                attr.fields(Service).project_id,
                attr.fields(Service).init_input,
                attr.fields(Service).module_name,
                attr.fields(Service).bot,
                attr.fields(Service).package_id,
                attr.fields(Service).is_global,
                attr.fields(Service).use_user_jwt,
                attr.fields(Service).package_revision,
                attr.fields(Service).driver_id,
                attr.fields(Service).run_execution_as_process,
                attr.fields(Service).execution_timeout,
                attr.fields(Service).drain_time,
                attr.fields(Service).runtime,
                attr.fields(Service).queue_length_limit,
                attr.fields(Service).max_attempts,
                attr.fields(Service).on_reset,
                attr.fields(Service).created_at,
                attr.fields(Service).updated_at,
                attr.fields(Service).secrets,
                attr.fields(Service)._type,
                attr.fields(Service).mode,
                attr.fields(Service).metadata,
                attr.fields(Service).archive,
                attr.fields(Service).updated_by,
                attr.fields(Service).config,
                attr.fields(Service).settings,
                attr.fields(Service).app,
                attr.fields(Service).integrations,
                attr.fields(Service).org_id,
                attr.fields(Service).panels
            )
        )

        _json['projectId'] = self.project_id
        _json['orgId'] = self.org_id
        _json['packageId'] = self.package_id
        _json['initParams'] = self.init_input
        _json['moduleName'] = self.module_name
        _json['botUserName'] = self.bot
        _json['useUserJwt'] = self.use_user_jwt
        _json['global'] = self.is_global
        _json['driverId'] = self.driver_id
        _json['packageRevision'] = self.package_revision
        _json['runExecutionAsProcess'] = self.run_execution_as_process
        _json['executionTimeout'] = self.execution_timeout
        _json['drainTime'] = self.drain_time
        _json['onReset'] = self.on_reset
        _json['createdAt'] = self.created_at
        _json['updatedAt'] = self.updated_at

        if self.updated_by is not None:
            _json['updatedBy'] = self.updated_by

        if self.panels is not None:
            _json['panels'] = self.panels

        if self.max_attempts is not None:
            _json['maxAttempts'] = self.max_attempts

        if self.is_global is not None:
            _json['global'] = self.is_global

        if self.runtime:
            _json['runtime'] = self.runtime if isinstance(self.runtime, dict) else self.runtime.to_json()

        if self.queue_length_limit is not None:
            _json['queueLengthLimit'] = self.queue_length_limit

        if self.secrets is not None:
            _json['secrets'] = self.secrets

        if self._type is not None:
            _json['type'] = self._type

        if self.mode:
            _json['mode'] = self.mode

        if self.metadata:
            _json['metadata'] = self.metadata

        if self.archive is not None:
            _json['archive'] = self.archive

        if self.config is not None:
            _json['config'] = self.config

        if self.settings is not None:
            _json['settings'] = self.settings

        if self.app is not None:
            _json['app'] = self.app

        if self.integrations is not None:
            _json['integrations'] = self.integrations

        return _json

    def update(self, force=False):
        """
        Update Service changes to platform

        :param bool force: force update
        :return: Service entity
        :rtype: dtlpy.entities.service.Service
        """
        return self.services.update(service=self, force=force)

    def delete(self, force: bool = False):
        """
        Delete Service object

        :return: True
        :rtype: bool
        """
        return self.services.delete(service_id=self.id, force=force)

    def status(self):
        """
        Get Service status

        :return: status json
        :rtype: dict
        """
        return self.services.status(service_id=self.id)

    def log(self,
            size=None,
            checkpoint=None,
            start=None,
            end=None,
            follow=False,
            text=None,
            execution_id=None,
            function_name=None,
            replica_id=None,
            system=False,
            view=True,
            until_completed=True,
            model_id: str = None,
            model_operation: str = None,
            ):
        """
        Get service logs

        :param int size: size
        :param dict checkpoint: the information from the lst point checked in the service
        :param str start: iso format time
        :param str end: iso format time
        :param bool follow: if true, keep stream future logs
        :param str text: text
        :param str execution_id: execution id
        :param str function_name: function name
        :param str replica_id: replica id
        :param bool system: system
        :param bool view: if true, print out all the logs
        :param bool until_completed: wait until completed
        :param str model_id: model id
        :param str model_operation: model operation action
        :return: ServiceLog entity
        :rtype: ServiceLog

        **Example**:

        .. code-block:: python

            service_log = service.log()
        """
        return self.services.log(service=self,
                                 size=size,
                                 checkpoint=checkpoint,
                                 start=start,
                                 end=end,
                                 follow=follow,
                                 execution_id=execution_id,
                                 function_name=function_name,
                                 replica_id=replica_id,
                                 system=system,
                                 text=text,
                                 view=view,
                                 until_completed=until_completed,
                                 model_id=model_id,
                                 model_operation=model_operation)

    def open_in_web(self):
        """
        Open the service in web platform

        :return:
        """
        parsed_url = urlsplit(self.platform_url)
        base_url = parsed_url.scheme + "://" + parsed_url.netloc
        url = '{}/projects/{}/services/{}'.format(base_url, self.project_id, self.id)
        self._client_api._open_in_web(url=url)

    def checkout(self):
        """
        Checkout

        :return:
        """
        return self.services.checkout(service=self)

    def pause(self):
        """
        pause

        :return:
        """
        return self.services.pause(service_id=self.id)

    def resume(self):
        """
        resume

        :return:
        """
        return self.services.resume(service_id=self.id)

    def execute(
            self,
            execution_input=None,
            function_name=None,
            resource=None,
            item_id=None,
            dataset_id=None,
            annotation_id=None,
            project_id=None,
            sync=False,
            stream_logs=True,
            return_output=True
    ):
        """
        Execute a function on an existing service

        :param List[FunctionIO] or dict execution_input: input dictionary or list of FunctionIO entities
        :param str function_name: function name to run
        :param str resource: input type.
        :param str item_id: optional - item id as input to function
        :param str dataset_id: optional - dataset id as input to function
        :param str annotation_id: optional - annotation id as input to function
        :param str project_id: resource's project
        :param bool sync: if true, wait for function to end
        :param bool stream_logs: prints logs of the new execution. only works with sync=True
        :param bool return_output: if True and sync is True - will return the output directly
        :return: execution object
        :rtype: dtlpy.entities.execution.Execution

        **Example**:

        .. code-block:: python

            execution = service.execute(function_name='function_name', item_id='item_id', project_id='project_id')
        """
        execution = self.executions.create(sync=sync,
                                           execution_input=execution_input,
                                           function_name=function_name,
                                           resource=resource,
                                           item_id=item_id,
                                           dataset_id=dataset_id,
                                           annotation_id=annotation_id,
                                           stream_logs=stream_logs,
                                           project_id=project_id,
                                           return_output=return_output)
        return execution

    def execute_batch(self,
                      filters,
                      function_name: str = None,
                      execution_inputs: list = None,
                      wait=True
                      ):
        """
        Execute a function on an existing service

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a service.

        :param filters: Filters entity for a filtering before execute
        :param str function_name: function name to run
        :param List[FunctionIO] or dict execution_inputs: input dictionary or list of FunctionIO entities, that represent the extra inputs of the function
        :param bool wait: wait until create task finish
        :return: execution object
        :rtype: dtlpy.entities.execution.Execution

        **Example**:

        .. code-block:: python

            command = service.execute_batch(
                        execution_inputs=dl.FunctionIO(type=dl.PackageInputType.STRING, value='test', name='string'),
                        filters=dl.Filters(field='dir', values='/test', context={"datasets": [dataset.id]),
                        function_name='run')
        """
        execution = self.executions.create_batch(service_id=self.id,
                                                 execution_inputs=execution_inputs,
                                                 filters=filters,
                                                 function_name=function_name,
                                                 wait=wait)
        return execution

    def rerun_batch(self,
                    filters,
                    wait=True
                    ):
        """
        rerun a executions on an existing service

        **Prerequisites**: You must be in the role of an *owner* or *developer*. You must have a Filter.

        :param filters: Filters entity for a filtering before rerun
        :param bool wait: wait until create task finish
        :return: rerun command
        :rtype: dtlpy.entities.command.Command

        **Example**:

        .. code-block:: python

            command = service.executions.rerun_batch(
                        filters=dl.Filters(field='id', values=['executionId'], operator=dl.FiltersOperations.IN, resource=dl.FiltersResource.EXECUTION))
        """
        execution = self.executions.rerun_batch(service_id=self.id,
                                                filters=filters,
                                                wait=wait)
        return execution

    def activate_slots(
            self,
            project_id: str = None,
            task_id: str = None,
            dataset_id: str = None,
            org_id: str = None,
            user_email: str = None,
            slots=None,
            role=None,
            prevent_override: bool = True,
            visible: bool = True,
            icon: str = 'fas fa-magic',
            **kwargs
    ) -> object:
        """
        Activate service slots

        :param str project_id: project id
        :param str task_id: task id
        :param str dataset_id: dataset id
        :param str org_id: org id
        :param str user_email: user email
        :param list slots: list of entities.PackageSlot
        :param str role: user role MemberOrgRole.ADMIN, MemberOrgRole.owner, MemberOrgRole.MEMBER, MemberOrgRole.WORKER
        :param bool prevent_override: True to prevent override
        :param bool visible: visible
        :param str icon: icon
        :param kwargs: all additional arguments
        :return: list of user setting for activated slots
        :rtype: list

        **Example**:

        .. code-block:: python

            setting = service.activate_slots(project_id='project_id',
                                    slots=List[entities.PackageSlot],
                                    icon='fas fa-magic')
        """
        return self.services.activate_slots(
            service=self,
            project_id=project_id,
            task_id=task_id,
            dataset_id=dataset_id,
            org_id=org_id,
            user_email=user_email,
            slots=slots,
            role=role,
            prevent_override=prevent_override,
            visible=visible,
            icon=icon,
            **kwargs
        )

    def restart(self, replica_name: str = None):
        """
        Restart service

        :param str replica_name: replica name
        :return: True
        :rtype: bool
        """
        return self.services.restart(service=self, replica_name=replica_name)


class KubernetesAutoscalerType(str, Enum):
    """ The Service Autoscaler Type (RABBITMQ, CPU).

    .. list-table::
       :widths: 15 150
       :header-rows: 1

       * - State
         - Description
       * - RABBITMQ
         - Service Autoscaler based on service queue length
       * - CPU
         - Service Autoscaler based on service CPU usage
       * - RPS
            - Service Autoscaler based on service RPS
    """
    RABBITMQ = 'rabbitmq'
    CPU = 'cpu'
    RPS = 'rps'


# added this class to avoid breaking changes after fixing a spelling mistake in KubernetesAutoscalerType
class KubernetesAutuscalerTypeMeta(type):
    def __getattribute__(cls, item):
        if hasattr(KubernetesAutoscalerType, item):
            return getattr(KubernetesAutoscalerType, item)
        else:
            raise AttributeError(f"KubernetesAutuscalerType has no attribute '{item}'")


class KubernetesAutoscaler(entities.BaseEntity):
    MIN_REPLICA_DEFAULT = 0
    MAX_REPLICA_DEFAULT = 1
    AUTOSCALER_TYPE_DEFAULT = KubernetesAutoscalerType.RABBITMQ

    def __init__(self,
                 autoscaler_type: KubernetesAutoscalerType.RABBITMQ = AUTOSCALER_TYPE_DEFAULT,
                 min_replicas=MIN_REPLICA_DEFAULT,
                 max_replicas=MAX_REPLICA_DEFAULT,
                 cooldown_period=None,
                 polling_interval=None,
                 **kwargs):
        self.autoscaler_type = kwargs.get('type', autoscaler_type)
        self.min_replicas = kwargs.get('minReplicas', min_replicas)
        self.max_replicas = kwargs.get('maxReplicas', max_replicas)
        self.cooldown_period = kwargs.get('cooldownPeriod', cooldown_period)
        self.polling_interval = kwargs.get('pollingInterval', polling_interval)

    def to_json(self):
        _json = {
            'type': self.autoscaler_type,
            'minReplicas': self.min_replicas,
            'maxReplicas': self.max_replicas
        }

        if self.cooldown_period is not None:
            _json['cooldownPeriod'] = self.cooldown_period

        if self.polling_interval is not None:
            _json['pollingInterval'] = self.polling_interval

        return _json


class KubernetesRabbitmqAutoscaler(KubernetesAutoscaler):
    QUEUE_LENGTH_DEFAULT = 1000

    def __init__(self,
                 min_replicas=KubernetesAutoscaler.MIN_REPLICA_DEFAULT,
                 max_replicas=KubernetesAutoscaler.MAX_REPLICA_DEFAULT,
                 queue_length=QUEUE_LENGTH_DEFAULT,
                 cooldown_period=None,
                 polling_interval=None,
                 **kwargs):
        super().__init__(min_replicas=min_replicas,
                         max_replicas=max_replicas,
                         autoscaler_type=KubernetesAutoscalerType.RABBITMQ,
                         cooldown_period=cooldown_period,
                         polling_interval=polling_interval, **kwargs)
        self.queue_length = kwargs.get('queueLength', queue_length)

    def to_json(self):
        _json = super().to_json()
        _json['queueLength'] = self.queue_length
        return _json


class KubernetesRPSAutoscaler(KubernetesAutoscaler):
    THRESHOLD_DEFAULT = 10
    RATE_SECONDS_DEFAULT = 30

    def __init__(self,
                 min_replicas=KubernetesAutoscaler.MIN_REPLICA_DEFAULT,
                 max_replicas=KubernetesAutoscaler.MAX_REPLICA_DEFAULT,
                 threshold=THRESHOLD_DEFAULT,
                 rate_seconds=RATE_SECONDS_DEFAULT,
                 cooldown_period=None,
                 polling_interval=None,
                 **kwargs):
        super().__init__(min_replicas=min_replicas,
                         max_replicas=max_replicas,
                         autoscaler_type=KubernetesAutoscalerType.RPS,
                         cooldown_period=cooldown_period,
                         polling_interval=polling_interval, **kwargs)
        self.threshold = kwargs.get('threshold', threshold)
        self.rate_seconds = kwargs.get('rateSeconds', rate_seconds)

    def to_json(self):
        _json = super().to_json()
        _json['rateSeconds'] = self.rate_seconds
        _json['threshold'] = self.threshold
        return _json



import logging
import traceback
import urllib.parse

from .. import entities, miscellaneous, repositories, exceptions, _api_reference
from ..services.api_client import ApiClient

logger = logging.getLogger(name='dtlpy')


class Recipes:
    """
    Recipes Repository

    The Recipes class allows you to manage recipes and their properties.
    For more information on Recipes, see our `documentation <https://dataloop.ai/docs/ontology>`_ and `developers' documentation <https://developers.dataloop.ai/tutorials/recipe_and_ontology/recipe/chapter/>`_.
    """

    def __init__(self,
                 client_api: ApiClient,
                 dataset: entities.Dataset = None,
                 project: entities.Project = None,
                 project_id: str = None):
        self._client_api = client_api
        self._dataset = dataset
        self._project = project
        self._project_id = project_id
        if project_id is None and project is not None:
            self._project_id = project.id

    ############
    # entities #
    ############
    @property
    def platform_url(self):
        if self._project_id is None:
            project_id = self.dataset.project.id
        else:
            project_id = self._project_id
        return self._client_api._get_resource_url("projects/{}/recipes".format(project_id))

    @property
    def project(self) -> entities.Project:
        if self._project is None:
            if self._project_id is None:
                if self._dataset is None:
                    raise exceptions.PlatformException(
                        error='2001',
                        message='Missing "Project". need to set a Project entity or use project.recipes repository'
                    )
                else:
                    self._project = self._dataset.project
                    self._project_id = self._project.id
            else:
                self._project = repositories.Projects(client_api=self._client_api).get(project_id=self._project_id)
                self._project_id = self._project.id

        assert isinstance(self._project, entities.Project)
        return self._project

    @project.setter
    def project(self, project: entities.Project):
        if not isinstance(project, entities.Project):
            raise ValueError('Must input a valid Project entity')
        self._project = project
        self._project_id = project.id

    @property
    def dataset(self) -> entities.Dataset:
        if self._dataset is None:
            raise exceptions.PlatformException(
                error='2001',
                message='Missing "dataset". need to set a Dataset entity or use dataset.recipes repository')
        assert isinstance(self._dataset, entities.Dataset)
        return self._dataset

    @dataset.setter
    def dataset(self, dataset: entities.Dataset):
        if not isinstance(dataset, entities.Dataset):
            raise ValueError('Must input a valid Dataset entity')
        self._dataset = dataset

    ###########
    # methods #
    ###########
    @_api_reference.add(path='/recipes', method='post')
    def create(self,
               project_ids=None,
               ontology_ids=None,
               labels=None,
               recipe_name=None,
               attributes=None,
               annotation_instruction_file=None
               ) -> entities.Recipe:
        """
        Create a new Recipe.
        Note: If the param ontology_ids is None, an ontology will be created first.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str project_ids: project ids
        :param str or list ontology_ids: ontology ids
        :param labels: labels
        :param str recipe_name: recipe name
        :param attributes: attributes
        :param str annotation_instruction_file: file path or url of the recipe instruction
        :return: Recipe entity
        :rtype: dtlpy.entities.recipe.Recipe

        **Example**:

        .. code-block:: python

            dataset.recipes.create(recipe_name='My Recipe', labels=labels))
        """
        if labels is None:
            labels = list()
        if attributes is None:
            attributes = list()
        if project_ids is None:
            if self._dataset is not None:
                project_ids = [self._dataset.project.id]
            else:
                # get from cache
                project = self._client_api.state_io.get('project')
                if project is not None:
                    # build entity from json
                    p = entities.Project.from_json(_json=project, client_api=self._client_api)
                    project_ids = [p.id]
                else:
                    # get from self.project property
                    try:
                        project_ids = [self.project.id]
                    except exceptions.PlatformException:
                        raise exceptions.PlatformException('Must provide project_ids')
        if ontology_ids is None:
            ontolgies = repositories.Ontologies(client_api=self._client_api,
                                                recipe=None)
            ontology = ontolgies.create(labels=labels,
                                        project_ids=project_ids,
                                        attributes=attributes)
            ontology_ids = [ontology.id]
        elif not isinstance(ontology_ids, list):
            ontology_ids = [ontology_ids]
        if recipe_name is None:
            recipe_name = self._dataset.name + " Default Recipe" if self._dataset is not None else "Default Recipe"
        payload = {'title': recipe_name,
                   'projectIds': project_ids,
                   'ontologyIds': ontology_ids,
                   'uiSettings': {
                       "allowObjectIdAutoAssign": True,
                       "studioV2App": True
                   }}
        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/recipes',
                                                         json_req=payload)
        if success:
            recipe = entities.Recipe.from_json(client_api=self._client_api,
                                               _json=response.json(),
                                               dataset=self._dataset)
            if annotation_instruction_file:
                recipe.add_instruction(annotation_instruction_file=annotation_instruction_file)
        else:
            logger.error('Failed to create Recipe')
            raise exceptions.PlatformException(response)

        if self._dataset is not None:
            self._dataset.switch_recipe(recipe.id)
        return recipe

    def list(self, filters: entities.Filters = None) -> miscellaneous.List[entities.Recipe]:
        """
        List recipes for a dataset.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.filters.Filters filters: Filters entity or a dictionary containing filters parameters
        :return: list of all recipes
        :retype: list

        **Example**:

        .. code-block:: python

            dataset.recipes.list()
        """
        if self._dataset is not None:
            try:
                recipes = [recipe_id for recipe_id in self._dataset.metadata['system']['recipes']]
            except KeyError:
                recipes = list()

            pool = self._client_api.thread_pools(pool_name='entity.create')
            jobs = [None for _ in range(len(recipes))]
            for i_recipe, recipe_id in enumerate(recipes):
                jobs[i_recipe] = pool.submit(self._protected_get, **{'recipe_id': recipe_id})

            # get all results
            results = [j.result() for j in jobs]
            # log errors
            _ = [logger.warning(r[1]) for r in results if r[0] is False]
            # return good jobs
            recipes = miscellaneous.List([r[1] for r in results if r[0] is True])
        elif self._project_id is not None:
            if filters is None:
                filters = entities.Filters(resource=entities.FiltersResource.RECIPE)
            # assert type filters
            elif not isinstance(filters, entities.Filters):
                raise exceptions.PlatformException(error='400',
                                                   message='Unknown filters type: {!r}'.format(type(filters)))
            if filters.resource != entities.FiltersResource.RECIPE:
                raise exceptions.PlatformException(
                    error='400',
                    message='Filters resource must to be FiltersResource.RECIPE. Got: {!r}'.format(filters.resource))
            if not filters.has_field('projects'):
                filters.add(field='projects', values=[self._project_id])

            recipes = entities.PagedEntities(items_repository=self,
                                             filters=filters,
                                             page_offset=filters.page,
                                             page_size=filters.page_size,
                                             project_id=self._project_id,
                                             client_api=self._client_api)
            recipes.get_page()
        else:
            raise exceptions.PlatformException('400', 'Must have project or dataset entity in repository')

        return recipes

    def _list(self, filters: entities.Filters):
        url = filters.generate_url_query_params('/recipes')
        encoded_url = urllib.parse.quote(url, safe='/:?=&')
        # request
        success, response = self._client_api.gen_request(req_type='get', path=encoded_url)
        if not success:
            raise exceptions.PlatformException(response)
        return response.json()

    def _build_entities_from_response(self, response_items) -> miscellaneous.List[entities.Recipe]:
        pool = self._client_api.thread_pools(pool_name='entity.create')
        jobs = [None for _ in range(len(response_items))]
        # return triggers list
        for i_rec, rec in enumerate(response_items):
            jobs[i_rec] = pool.submit(entities.Recipe._protected_from_json,
                                      **{'client_api': self._client_api,
                                         '_json': rec,
                                         'project': self._project,
                                         'dataset': self._dataset})

        # get all results
        results = [j.result() for j in jobs]
        # log errors
        _ = [logger.warning(r[1]) for r in results if r[0] is False]
        # return good jobs
        recipes = miscellaneous.List([r[1] for r in results if r[0] is True])
        return recipes

    def _protected_get(self, recipe_id):
        """
        Same as get but with try-except to catch if error
        :param recipe_id:
        :return:
        """
        try:
            recipe = self.get(recipe_id=recipe_id)
            status = True
        except Exception:
            recipe = traceback.format_exc()
            status = False
        return status, recipe

    @_api_reference.add(path='/recipes/{id}', method='get')
    def get(self, recipe_id: str) -> entities.Recipe:
        """
        Get a Recipe object to use in your code.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str recipe_id: recipe id
        :return: Recipe object
        :rtype: dtlpy.entities.recipe.Recipe

        **Example**:

        .. code-block:: python

            dataset.recipes.get(recipe_id='recipe_id')
        """
        success, response = self._client_api.gen_request(req_type='get',
                                                         path='/recipes/%s' % recipe_id)
        if success:
            recipe = entities.Recipe.from_json(client_api=self._client_api,
                                               _json=response.json(),
                                               project=self._project,
                                               dataset=self._dataset)
        else:
            logger.error('Unable to get info from recipe. Recipe_id id: {}'.format(recipe_id))
            raise exceptions.PlatformException(response)

        return recipe

    def open_in_web(self,
                    recipe: entities.Recipe = None,
                    recipe_id: str = None):
        """
        Open the recipe in web platform.

        **Prerequisites**: All users.

        :param dtlpy.entities.recipe.Recipe recipe: recipe entity
        :param str recipe_id: recipe id

        **Example**:

        .. code-block:: python

            dataset.recipes.open_in_web(recipe_id='recipe_id')
        """
        if recipe is not None:
            recipe.open_in_web()
        elif recipe_id is not None:
            self._client_api._open_in_web(url=self.platform_url + '/' + str(recipe_id))
        else:
            self._client_api._open_in_web(url=self.platform_url)

    @_api_reference.add(path='/recipes/{id}', method='delete')
    def delete(self, recipe_id: str, force: bool = False):
        """
        Delete recipe from platform.
        
        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param str recipe_id: recipe id
        :param bool force: force delete recipe
        :return: True if success
        :rtype: bool

        **Example**:

        .. code-block:: python

            dataset.recipes.delete(recipe_id='recipe_id')
        """
        path = '/recipes/{}'.format(recipe_id)
        if force:
            path += '?force=true'
        success, response = self._client_api.gen_request(req_type='delete',
                                                         path=path)
        if not success:
            raise exceptions.PlatformException(response)
        logger.info('Recipe id {} deleted successfully'.format(recipe_id))
        return True

    @_api_reference.add(path='/recipes/{id}', method='patch')
    def update(self, recipe: entities.Recipe, system_metadata=False) -> entities.Recipe:
        """
        Update recipe.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

        :param dtlpy.entities.recipe.Recipe recipe: Recipe object
        :param bool system_metadata: True, if you want to change metadata system
        :return: Recipe object
        :rtype: dtlpy.entities.recipe.Recipe

        **Example**:

        .. code-block:: python

            dataset.recipes.update(recipe='recipe_entity')
        """
        url_path = '/recipes/%s' % recipe.id
        if system_metadata:
            url_path += '?system=true'
        success, response = self._client_api.gen_request(req_type='patch',
                                                         path=url_path,
                                                         json_req=recipe.to_json())
        if success:
            return entities.Recipe.from_json(client_api=self._client_api, _json=response.json(), dataset=self._dataset)
        else:
            logger.error('Error while updating item:')
            raise exceptions.PlatformException(response)

    @_api_reference.add(path='/recipes/{id}/clone', method='post')
    def clone(self,
              recipe: entities.Recipe = None,
              recipe_id: str = None,
              shallow: bool = False):
        """
        Clone recipe.

        **Prerequisites**: You must be in the role of an *owner* or *developer*.

       :param dtlpy.entities.recipe.Recipe recipe: Recipe object
       :param str recipe_id: Recipe id
       :param bool shallow: If True, link to existing ontology, clones all ontologies that are linked to the recipe as well
       :return: Cloned ontology object
       :rtype: dtlpy.entities.recipe.Recipe

       **Example**:

        .. code-block:: python

            dataset.recipes.clone(recipe_id='recipe_id')
       """
        if recipe is None and recipe_id is None:
            raise exceptions.PlatformException('400', 'Must provide recipe or recipe_id')
        if recipe_id is None:
            if not isinstance(recipe, entities.Recipe):
                raise exceptions.PlatformException('400', 'Recipe must me entities.Recipe type')
            else:
                recipe_id = recipe.id

        payload = {'shallow': shallow}

        success, response = self._client_api.gen_request(req_type='post',
                                                         path='/recipes/{}/clone'.format(recipe_id),
                                                         json_req=payload)
        if success:
            recipe = entities.Recipe.from_json(client_api=self._client_api,
                                               _json=response.json())
        else:
            logger.error('Failed to clone Recipe')
            raise exceptions.PlatformException(response)

        assert isinstance(recipe, entities.Recipe)
        logger.debug('Recipe has been cloned successfully. recipe id: {}'.format(recipe.id))

        return recipe
